import base64 as _base64
import hashlib as _hashlib
import os as _os
import re as _re
import webbrowser as _webbrowser
from multiprocessing import Process as _Process
from multiprocessing import Queue as _Queue

import keyring as _keyring
import requests as _requests

try:  # Python 3.5+
    from http import HTTPStatus as _StatusCodes
except ImportError:
    try:  # Python 3
        from http import client as _StatusCodes
    except ImportError:  # Python 2
        import httplib as _StatusCodes
try:  # Python 3
    import http.server as _BaseHTTPServer
except ImportError:  # Python 2
    import BaseHTTPServer as _BaseHTTPServer

try:  # Python 3
    import urllib.parse as _urlparse
    from urllib.parse import urlencode as _urlencode
except ImportError:  # Python 2
    from urllib import urlencode as _urlencode

    import urlparse as _urlparse

_code_verifier_length = 64
_random_seed_length = 40
_utf_8 = "utf-8"


# Identifies the service used for storing passwords in keyring
_keyring_service_name = "flyteauth"
# Identifies the key used for storing and fetching from keyring. In our case, instead of a username as the keyring docs
# suggest, we are storing a user's oidc.
_keyring_access_token_storage_key = "access_token"
_keyring_refresh_token_storage_key = "refresh_token"


def _generate_code_verifier():
    """
    Generates a 'code_verifier' as described in https://tools.ietf.org/html/rfc7636#section-4.1
    Adapted from https://github.com/openstack/deb-python-oauth2client/blob/master/oauth2client/_pkce.py.
    :return str:
    """
    code_verifier = _base64.urlsafe_b64encode(_os.urandom(_code_verifier_length)).decode(_utf_8)
    # Eliminate invalid characters.
    code_verifier = _re.sub(r"[^a-zA-Z0-9_\-.~]+", "", code_verifier)
    if len(code_verifier) < 43:
        raise ValueError("Verifier too short. number of bytes must be > 30.")
    elif len(code_verifier) > 128:
        raise ValueError("Verifier too long. number of bytes must be < 97.")
    return code_verifier


def _generate_state_parameter():
    state = _base64.urlsafe_b64encode(_os.urandom(_random_seed_length)).decode(_utf_8)
    # Eliminate invalid characters.
    code_verifier = _re.sub("[^a-zA-Z0-9-_.,]+", "", state)
    return code_verifier


def _create_code_challenge(code_verifier):
    """
    Adapted from https://github.com/openstack/deb-python-oauth2client/blob/master/oauth2client/_pkce.py.
    :param str code_verifier: represents a code verifier generated by generate_code_verifier()
    :return str: urlsafe base64-encoded sha256 hash digest
    """
    code_challenge = _hashlib.sha256(code_verifier.encode(_utf_8)).digest()
    code_challenge = _base64.urlsafe_b64encode(code_challenge).decode(_utf_8)
    # Eliminate invalid characters
    code_challenge = code_challenge.replace("=", "")
    return code_challenge


class AuthorizationCode(object):
    def __init__(self, code, state):
        self._code = code
        self._state = state

    @property
    def code(self):
        return self._code

    @property
    def state(self):
        return self._state


class OAuthCallbackHandler(_BaseHTTPServer.BaseHTTPRequestHandler):
    """
    A simple wrapper around BaseHTTPServer.BaseHTTPRequestHandler that handles a callback URL that accepts an
    authorization token.
    """

    def do_GET(self):
        url = _urlparse.urlparse(self.path)
        if url.path == self.server.redirect_path:
            self.send_response(_StatusCodes.OK)
            self.end_headers()
            self.handle_login(dict(_urlparse.parse_qsl(url.query)))
        else:
            self.send_response(_StatusCodes.NOT_FOUND)

    def handle_login(self, data):
        self.server.handle_authorization_code(AuthorizationCode(data["code"], data["state"]))


class OAuthHTTPServer(_BaseHTTPServer.HTTPServer):
    """
    A simple wrapper around the BaseHTTPServer.HTTPServer implementation that binds an authorization_client for handling
    authorization code callbacks.
    """

    def __init__(
        self,
        server_address,
        RequestHandlerClass,
        bind_and_activate=True,
        redirect_path=None,
        queue=None,
    ):
        _BaseHTTPServer.HTTPServer.__init__(self, server_address, RequestHandlerClass, bind_and_activate)
        self._redirect_path = redirect_path
        self._auth_code = None
        self._queue = queue

    @property
    def redirect_path(self):
        return self._redirect_path

    def handle_authorization_code(self, auth_code):
        self._queue.put(auth_code)


class Credentials(object):
    def __init__(self, access_token=None):
        self._access_token = access_token

    @property
    def access_token(self):
        return self._access_token


class AuthorizationClient(object):
    def __init__(self, auth_endpoint=None, token_endpoint=None, client_id=None, redirect_uri=None):
        self._auth_endpoint = auth_endpoint
        self._token_endpoint = token_endpoint
        self._client_id = client_id
        self._redirect_uri = redirect_uri
        self._code_verifier = _generate_code_verifier()
        code_challenge = _create_code_challenge(self._code_verifier)
        self._code_challenge = code_challenge
        state = _generate_state_parameter()
        self._state = state
        self._credentials = None
        self._refresh_token = None
        self._headers = {"content-type": "application/x-www-form-urlencoded"}
        self._expired = False

        self._params = {
            "client_id": client_id,  # This must match the Client ID of the OAuth application.
            "response_type": "code",  # Indicates the authorization code grant
            "scope": "openid offline_access",  # ensures that the /token endpoint returns an ID and refresh token
            # callback location where the user-agent will be directed to.
            "redirect_uri": self._redirect_uri,
            "state": state,
            "code_challenge": code_challenge,
            "code_challenge_method": "S256",
        }

        # Prefer to use already-fetched token values when they've been set globally.
        self._refresh_token = _keyring.get_password(_keyring_service_name, _keyring_refresh_token_storage_key)
        access_token = _keyring.get_password(_keyring_service_name, _keyring_access_token_storage_key)
        if access_token:
            self._credentials = Credentials(access_token=access_token)
            return

        # In the absence of globally-set token values, initiate the token request flow
        q = _Queue()
        # First prepare the callback server in the background
        server = self._create_callback_server(q)
        server_process = _Process(target=server.handle_request)
        server_process.start()

        # Send the call to request the authorization code
        self._request_authorization_code()

        # Request the access token once the auth code has been received.
        auth_code = q.get()
        server_process.terminate()
        self.request_access_token(auth_code)

    def _create_callback_server(self, q):
        server_url = _urlparse.urlparse(self._redirect_uri)
        server_address = (server_url.hostname, server_url.port)
        return OAuthHTTPServer(server_address, OAuthCallbackHandler, redirect_path=server_url.path, queue=q)

    def _request_authorization_code(self):
        scheme, netloc, path, _, _, _ = _urlparse.urlparse(self._auth_endpoint)
        query = _urlencode(self._params)
        endpoint = _urlparse.urlunparse((scheme, netloc, path, None, query, None))
        _webbrowser.open_new_tab(endpoint)

    def _initialize_credentials(self, auth_token_resp):

        """
        The auth_token_resp body is of the form:
        {
          "access_token": "foo",
          "refresh_token": "bar",
          "id_token": "baz",
          "token_type": "Bearer"
        }
        """
        response_body = auth_token_resp.json()
        if "access_token" not in response_body:
            raise ValueError('Expected "access_token" in response from oauth server')
        if "refresh_token" in response_body:
            self._refresh_token = response_body["refresh_token"]

        access_token = response_body["access_token"]
        refresh_token = response_body["refresh_token"]

        _keyring.set_password(_keyring_service_name, _keyring_access_token_storage_key, access_token)
        _keyring.set_password(_keyring_service_name, _keyring_refresh_token_storage_key, refresh_token)
        self._credentials = Credentials(access_token=access_token)

    def request_access_token(self, auth_code):
        if self._state != auth_code.state:
            raise ValueError("Unexpected state parameter [{}] passed".format(auth_code.state))
        self._params.update(
            {"code": auth_code.code, "code_verifier": self._code_verifier, "grant_type": "authorization_code"}
        )
        resp = _requests.post(
            url=self._token_endpoint,
            data=self._params,
            headers=self._headers,
            allow_redirects=False,
        )
        if resp.status_code != _StatusCodes.OK:
            # TODO: handle expected (?) error cases:
            #  https://auth0.com/docs/flows/guides/device-auth/call-api-device-auth#token-responses
            raise Exception(
                "Failed to request access token with response: [{}] {}".format(resp.status_code, resp.content)
            )
        self._initialize_credentials(resp)

    def refresh_access_token(self):
        if self._refresh_token is None:
            raise ValueError("no refresh token available with which to refresh authorization credentials")

        resp = _requests.post(
            url=self._token_endpoint,
            data={"grant_type": "refresh_token", "client_id": self._client_id, "refresh_token": self._refresh_token},
            headers=self._headers,
            allow_redirects=False,
        )
        if resp.status_code != _StatusCodes.OK:
            self._expired = True
            # In the absence of a successful response, assume the refresh token is expired. This should indicate
            # to the caller that the AuthorizationClient is defunct and a new one needs to be re-initialized.

            _keyring.delete_password(_keyring_service_name, _keyring_access_token_storage_key)
            _keyring.delete_password(_keyring_service_name, _keyring_refresh_token_storage_key)
            return
        self._initialize_credentials(resp)

    @property
    def credentials(self):
        """
        :return flytekit.clis.auth.auth.Credentials:
        """
        return self._credentials

    @property
    def expired(self):
        """
        :return bool:
        """
        return self._expired



"""
=====================
Unit Testing
=====================

.. currentmodule:: flytekit.testing

The imports exposed in this package will help you unit test your Flyte tasks. These are particularly helpful when
testing workflows that contain tasks that cannot run locally (a Hive task for instance).

.. autosummary::
   :toctree: generated/

   patch - A decorator similar to the regular one you're probably used to
   task_mock - Non-decorative function

"""

from flytekit.common.tasks.sdk_runnable import SecretsManager
from flytekit.core.testing import patch, task_mock



import random as _random

random = _random.Random()
"""
An instance of the global random number generator used by flytekit.  Flytekit maintains it's own random instance
to ensure that calls to random.seed(...) do not affect the pseudo-random behavior of flytekit. This random should be
used by flytekit components in all cases where random.random would have been used. Components who want additional
protections for their random number generator might also maintain their own separate random instance.
"""


def seed_flyte_random(seed):
    """
    If one wants to influence the pseudo-random behavior of flytekit, this function can be used to seed the flytekit
    generator. It is not recommended that this be done as lack of entropy between jobs can result in overwriting data
    created at random locations.

    Currently, this is used by flytekit to create entropy in low entropy situations (such as Array Jobs) where the job
    index can be used as a seed to ensure sibling jobs do not have random collisions.
    :param Union[Text,int,bytes] seed:
    """
    global random
    random = _random.Random(seed)



import click as _click

CTX_PROJECT = "project"
CTX_DOMAIN = "domain"
CTX_VERSION = "version"
CTX_TEST = "test"
CTX_PACKAGES = "pkgs"
CTX_NOTIFICATIONS = "notifications"


project_option = _click.option(
    "-p",
    "--project",
    required=True,
    type=str,
    help="Flyte project to use. You can have more than one project per repo",
)
domain_option = _click.option(
    "-d",
    "--domain",
    required=True,
    type=str,
    help="This is usually development, staging, or production",
)
version_option = _click.option(
    "-v",
    "--version",
    required=False,
    type=str,
    help="This is the version to apply globally for this context",
)



INPUT_FILE_NAME = "inputs.pb"
OUTPUT_FILE_NAME = "outputs.pb"
FUTURES_FILE_NAME = "futures.pb"
ERROR_FILE_NAME = "error.pb"


class SdkTaskType(object):
    PYTHON_TASK = "python-task"
    DYNAMIC_TASK = "dynamic-task"
    CONTAINER_ARRAY_TASK = "container_array"
    SPARK_TASK = "spark"

    # Hive is multi-step operation:
    #    1. a generator task that generates hive-job to be executed by the operator. Generator task is called hive task
    #       for backward compatibility (Note: it is a "batch-task" with a different name)
    #    2. hive-job is the actual set of queries to be executed. This is called hive_job
    BATCH_HIVE_TASK = "batch_hive"
    HIVE_JOB = "hive"
    SIDECAR_TASK = "sidecar"
    SENSOR_TASK = "sensor-task"
    PRESTO_TASK = "presto"
    PYTORCH_TASK = "pytorch"
    TENSORFLOW_TASK = "tensorflow"
    # Raw container task is just a name, it defaults to using the regular container task (like python etc), but sets the data_config in the container
    RAW_CONTAINER_TASK = "raw-container"
    SAGEMAKER_TRAINING_JOB_TASK = "sagemaker_training_job_task"
    SAGEMAKER_CUSTOM_TRAINING_JOB_TASK = "sagemaker_custom_training_job_task"
    SAGEMAKER_HYPERPARAMETER_TUNING_JOB_TASK = "sagemaker_hyperparameter_tuning_job_task"


GLOBAL_INPUT_NODE_ID = ""

START_NODE_ID = "start-node"
END_NODE_ID = "end-node"


class CloudProvider(object):
    AWS = "aws"
    GCP = "gcp"
    LOCAL = "local"



import base64 as _base64
import logging as _logging

import requests as _requests

from flytekit.common.exceptions.user import FlyteAuthenticationException as _FlyteAuthenticationException
from flytekit.configuration.creds import CLIENT_CREDENTIALS_SECRET as _CREDENTIALS_SECRET

_utf_8 = "utf-8"


def get_secret():
    """
    This function will either read in the password from the file path given by the CLIENT_CREDENTIALS_SECRET_LOCATION
    config object, or from the environment variable using the CLIENT_CREDENTIALS_SECRET config object.
    :rtype: Text
    """
    secret = _CREDENTIALS_SECRET.get()
    if secret:
        return secret
    raise _FlyteAuthenticationException("No secret could be found")


def get_basic_authorization_header(client_id, client_secret):
    """
    This function transforms the client id and the client secret into a header that conforms with http basic auth.
    It joins the id and the secret with a : then base64 encodes it, then adds the appropriate text.
    :param Text client_id:
    :param Text client_secret:
    :rtype: Text
    """
    concated = "{}:{}".format(client_id, client_secret)
    return "Basic {}".format(_base64.b64encode(concated.encode(_utf_8)).decode(_utf_8))


def get_token(token_endpoint, authorization_header, scope):
    """
    :param Text token_endpoint:
    :param Text authorization_header: This is the value for the "Authorization" key. (eg 'Bearer abc123')
    :param Text scope:
    :rtype: (Text,Int) The first element is the access token retrieved from the IDP, the second is the expiration
            in seconds
    """
    headers = {
        "Authorization": authorization_header,
        "Cache-Control": "no-cache",
        "Accept": "application/json",
        "Content-Type": "application/x-www-form-urlencoded",
    }
    body = {
        "grant_type": "client_credentials",
    }
    if scope is not None:
        body["scope"] = scope
    response = _requests.post(token_endpoint, data=body, headers=headers)
    if response.status_code != 200:
        _logging.error("Non-200 ({}) received from IDP: {}".format(response.status_code, response.text))
        raise _FlyteAuthenticationException("Non-200 received from IDP")

    response = response.json()
    return response["access_token"], response["expires_in"]



import datetime as _datetime
import importlib as _importlib
import inspect as _inspect
import json as _json
import os as _os
import sys as _sys

import six as _six
from google.protobuf import json_format as _json_format
from google.protobuf import text_format as _text_format

from flytekit import __version__
from flytekit.bin import entrypoint as _entrypoint
from flytekit.common import constants as _constants
from flytekit.common import interface as _interface2
from flytekit.common.exceptions import scopes as _exception_scopes
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.tasks import output as _task_output
from flytekit.common.tasks import sdk_runnable as _sdk_runnable
from flytekit.common.tasks import spark_task as _spark_task
from flytekit.common.tasks import task as _base_tasks
from flytekit.common.types import helpers as _type_helpers
from flytekit.contrib.notebook.supported_types import notebook_types_map as _notebook_types_map
from flytekit.engines import loader as _engine_loader
from flytekit.models import interface as _interface
from flytekit.models import literals as _literal_models
from flytekit.models import task as _task_models
from flytekit.plugins import papermill as _pm
from flytekit.sdk.spark_types import SparkType as _spark_type
from flytekit.sdk.types import Types as _Types

OUTPUT_NOTEBOOK = "output_notebook"


def python_notebook(
    notebook_path="",
    inputs={},
    outputs={},
    cache_version="",
    retries=0,
    deprecated="",
    storage_request=None,
    cpu_request=None,
    gpu_request=None,
    memory_request=None,
    storage_limit=None,
    cpu_limit=None,
    gpu_limit=None,
    memory_limit=None,
    cache=False,
    timeout=None,
    environment=None,
    cls=None,
):
    """
    Decorator to create a Python Notebook Task definition.

    :rtype: SdkNotebookTask
    """
    return SdkNotebookTask(
        notebook_path=notebook_path,
        inputs=inputs,
        outputs=outputs,
        task_type=_constants.SdkTaskType.PYTHON_TASK,
        discovery_version=cache_version,
        retries=retries,
        deprecated=deprecated,
        storage_request=storage_request,
        cpu_request=cpu_request,
        gpu_request=gpu_request,
        memory_request=memory_request,
        storage_limit=storage_limit,
        cpu_limit=cpu_limit,
        gpu_limit=gpu_limit,
        memory_limit=memory_limit,
        discoverable=cache,
        timeout=timeout or _datetime.timedelta(seconds=0),
        environment=environment,
        custom={},
    )


class SdkNotebookTask(_base_tasks.SdkTask):

    """
    This class includes the additional logic for building a task that executes Notebooks.

    """

    def __init__(
        self,
        notebook_path,
        inputs,
        outputs,
        task_type,
        discovery_version,
        retries,
        deprecated,
        storage_request,
        cpu_request,
        gpu_request,
        memory_request,
        storage_limit,
        cpu_limit,
        gpu_limit,
        memory_limit,
        discoverable,
        timeout,
        environment,
        custom,
    ):

        if _os.path.isabs(notebook_path) is False:
            # Find absolute path for the notebook.
            task_module = _importlib.import_module(_find_instance_module())
            module_path = _os.path.dirname(task_module.__file__)
            notebook_path = _os.path.normpath(_os.path.join(module_path, notebook_path))

        self._notebook_path = notebook_path

        super(SdkNotebookTask, self).__init__(
            task_type,
            _task_models.TaskMetadata(
                discoverable,
                _task_models.RuntimeMetadata(
                    _task_models.RuntimeMetadata.RuntimeType.FLYTE_SDK,
                    __version__,
                    "notebook",
                ),
                timeout,
                _literal_models.RetryStrategy(retries),
                False,
                discovery_version,
                deprecated,
            ),
            _interface2.TypedInterface({}, {}),
            custom,
            container=self._get_container_definition(
                storage_request=storage_request,
                cpu_request=cpu_request,
                gpu_request=gpu_request,
                memory_request=memory_request,
                storage_limit=storage_limit,
                cpu_limit=cpu_limit,
                gpu_limit=gpu_limit,
                memory_limit=memory_limit,
                environment=environment,
            ),
        )
        # Add Inputs
        if inputs is not None:
            inputs(self)

        # Add outputs
        if outputs is not None:
            outputs(self)

        # Add a Notebook output as a Blob.
        self.interface.outputs.update(
            output_notebook=_interface.Variable(_Types.Blob.to_flyte_literal_type(), OUTPUT_NOTEBOOK)
        )

    def _validate_inputs(self, inputs):
        """
        :param dict[Text, flytekit.models.interface.Variable] inputs:  Input variables to validate
        :raises: flytekit.common.exceptions.user.FlyteValidationException
        """
        for k, v in _six.iteritems(inputs):
            sdk_type = _type_helpers.get_sdk_type_from_literal_type(v.type)
            if sdk_type not in _notebook_types_map.values():
                raise _user_exceptions.FlyteValidationException(
                    "Input Type '{}' not supported.  Only Primitives are supported for notebook.".format(sdk_type)
                )
        super(SdkNotebookTask, self)._validate_inputs(inputs)

    def _validate_outputs(self, outputs):
        """
        :param dict[Text, flytekit.models.interface.Variable] inputs:  Input variables to validate
        :raises: flytekit.common.exceptions.user.FlyteValidationException
        """

        # Add output_notebook as an implicit output to the task.

        for k, v in _six.iteritems(outputs):

            if k == OUTPUT_NOTEBOOK:
                raise ValueError(
                    "{} is a reserved output keyword. Please use a different output name.".format(OUTPUT_NOTEBOOK)
                )

            sdk_type = _type_helpers.get_sdk_type_from_literal_type(v.type)
            if sdk_type not in _notebook_types_map.values():
                raise _user_exceptions.FlyteValidationException(
                    "Output Type '{}' not supported.  Only Primitives are supported for notebook.".format(sdk_type)
                )
        super(SdkNotebookTask, self)._validate_outputs(outputs)

    @_exception_scopes.system_entry_point
    def add_inputs(self, inputs):
        """
        Adds the inputs to this task.  This can be called multiple times, but it will fail if an input with a given
        name is added more than once, a name collides with an output, or if the name doesn't exist as an arg name in
        the wrapped function.
        :param dict[Text, flytekit.models.interface.Variable] inputs: names and variables
        """
        self._validate_inputs(inputs)
        self.interface.inputs.update(inputs)

    @_exception_scopes.system_entry_point
    def add_outputs(self, outputs):
        """
        Adds the outputs to this task.  This can be called multiple times, but it will fail if an output with a given
        name is added more than once, a name collides with an output, or if the name doesn't exist as an arg name in
        the wrapped function.
        :param dict[Text, flytekit.models.interface.Variable] outputs: names and variables
        """
        self._validate_outputs(outputs)
        self.interface.outputs.update(outputs)

    @_exception_scopes.system_entry_point
    def unit_test(self, **input_map):
        """
        :param dict[Text, T] input_map: Python Std input from users.  We will cast these to the appropriate Flyte
            literals.
        :returns: Depends on the behavior of the specific task in the unit engine.
        """

        return (
            _engine_loader.get_engine("unit")
            .get_task(self)
            .execute(
                _type_helpers.pack_python_std_map_to_literal_map(
                    input_map,
                    {
                        k: _type_helpers.get_sdk_type_from_literal_type(v.type)
                        for k, v in _six.iteritems(self.interface.inputs)
                    },
                )
            )
        )

    @_exception_scopes.system_entry_point
    def local_execute(self, **input_map):
        """
        :param dict[Text, T] input_map: Python Std input from users.  We will cast these to the appropriate Flyte
            literals.
        :rtype: dict[Text, T]
        :returns: The output produced by this task in Python standard format.
        """
        return (
            _engine_loader.get_engine("local")
            .get_task(self)
            .execute(
                _type_helpers.pack_python_std_map_to_literal_map(
                    input_map,
                    {
                        k: _type_helpers.get_sdk_type_from_literal_type(v.type)
                        for k, v in _six.iteritems(self.interface.inputs)
                    },
                )
            )
        )

    @_exception_scopes.system_entry_point
    def execute(self, context, inputs):
        """
        :param flytekit.engines.common.EngineContext context:
        :param flytekit.models.literals.LiteralMap inputs:
        :rtype: dict[Text, flytekit.models.common.FlyteIdlEntity]
        :returns: This function must return a dictionary mapping 'filenames' to Flyte Interface Entities.  These
            entities will be used by the engine to pass data from node to node, populate metadata, etc. etc..  Each
            engine will have different behavior.  For instance, the Flyte engine will upload the entities to a remote
            working directory (with the names provided), which will in turn allow Flyte Propeller to push along the
            workflow.  Where as local engine will merely feed the outputs directly into the next node.
        """
        inputs_dict = _type_helpers.unpack_literal_map_to_sdk_python_std(
            inputs,
            {k: _type_helpers.get_sdk_type_from_literal_type(v.type) for k, v in _six.iteritems(self.interface.inputs)},
        )

        input_notebook_path = self._notebook_path
        # Execute Notebook via Papermill.
        output_notebook_path = input_notebook_path.split(".ipynb")[0] + "-out.ipynb"
        _pm.execute_notebook(input_notebook_path, output_notebook_path, parameters=inputs_dict)

        # Parse Outputs from Notebook.
        outputs = None
        with open(output_notebook_path) as json_file:
            data = _json.load(json_file)
            for p in data["cells"]:
                meta = p["metadata"]
                if "outputs" in meta["tags"]:
                    outputs = " ".join(p["outputs"][0]["data"]["text/plain"])

        if outputs is not None:
            dict = _literal_models._literals_pb2.LiteralMap()
            _text_format.Parse(outputs, dict)

        # Add output_notebook as an output to the task.
        output_notebook = _task_output.OutputReference(
            _type_helpers.get_sdk_type_from_literal_type(_Types.Blob.to_flyte_literal_type())
        )
        output_notebook.set(output_notebook_path)

        output_literal_map = _literal_models.LiteralMap.from_flyte_idl(dict)
        output_literal_map.literals[OUTPUT_NOTEBOOK] = output_notebook.sdk_value

        return {_constants.OUTPUT_FILE_NAME: output_literal_map}

    @property
    def container(self):
        """
        If not None, the target of execution should be a container.
        :rtype: Container
        """

        # Find task_name
        task_module = _importlib.import_module(self.instantiated_in)
        for k in dir(task_module):
            if getattr(task_module, k) is self:
                task_name = k
                break

        self._container._args = [
            "pyflyte-execute",
            "--task-module",
            self.instantiated_in,
            "--task-name",
            task_name,
            "--inputs",
            "{{.input}}",
            "--output-prefix",
            "{{.outputPrefix}}",
            "--raw-output-data-prefix",
            "{{.rawOutputDataPrefix}}",
        ]
        return self._container

    def _get_container_definition(
        self,
        storage_request=None,
        cpu_request=None,
        gpu_request=None,
        memory_request=None,
        storage_limit=None,
        cpu_limit=None,
        gpu_limit=None,
        memory_limit=None,
        environment=None,
        **kwargs
    ):
        """
        :param Text storage_request:
        :param Text cpu_request:
        :param Text gpu_request:
        :param Text memory_request:
        :param Text storage_limit:
        :param Text cpu_limit:
        :param Text gpu_limit:
        :param Text memory_limit:
        :param dict[Text, Text] environment:
        :rtype: flytekit.models.task.Container
        """

        storage_limit = storage_limit or storage_request
        cpu_limit = cpu_limit or cpu_request
        gpu_limit = gpu_limit or gpu_request
        memory_limit = memory_limit or memory_request

        resources = _sdk_runnable.SdkRunnableContainer.get_resources(
            storage_request, cpu_request, gpu_request, memory_request, storage_limit, cpu_limit, gpu_limit, memory_limit
        )

        return _sdk_runnable.SdkRunnableContainer(
            command=[],
            args=[],
            resources=resources,
            env=environment,
            config={},
        )


def spark_notebook(
    notebook_path,
    inputs={},
    outputs={},
    spark_conf=None,
    cache_version="",
    retries=0,
    deprecated="",
    cache=False,
    timeout=None,
    environment=None,
):
    """
    Decorator to create a Notebook spark task. This task will connect to a Spark cluster, configure the environment,
    and then execute the code within the notebook_path as the Spark driver program.
    """
    return SdkNotebookSparkTask(
        notebook_path=notebook_path,
        inputs=inputs,
        outputs=outputs,
        spark_conf=spark_conf,
        discovery_version=cache_version,
        retries=retries,
        deprecated=deprecated,
        discoverable=cache,
        timeout=timeout or _datetime.timedelta(seconds=0),
        environment=environment or {},
    )


def _find_instance_module():
    frame = _inspect.currentframe()
    while frame:
        if frame.f_code.co_name == "<module>":
            return frame.f_globals["__name__"]
        frame = frame.f_back
    return None


class SdkNotebookSparkTask(SdkNotebookTask):

    """
    This class includes the additional logic for building a task that executes Spark Notebooks.

    """

    def __init__(
        self,
        notebook_path,
        inputs,
        outputs,
        spark_conf,
        discovery_version,
        retries,
        deprecated,
        discoverable,
        timeout,
        environment=None,
    ):

        spark_exec_path = _os.path.abspath(_entrypoint.__file__)
        if spark_exec_path.endswith(".pyc"):
            spark_exec_path = spark_exec_path[:-1]

        if spark_conf is None:
            # Parse spark_conf from notebook if not set at task_level.
            with open(notebook_path) as json_file:
                data = _json.load(json_file)
                for p in data["cells"]:
                    meta = p["metadata"]
                    if "tags" in meta:
                        if "conf" in meta["tags"]:
                            sc_str = " ".join(p["source"])
                            ldict = {}
                            exec(sc_str, globals(), ldict)
                            spark_conf = ldict["spark_conf"]

            spark_job = _task_models.SparkJob(
                spark_conf=spark_conf,
                main_class="",
                spark_type=_spark_type.PYTHON,
                hadoop_conf={},
                application_file="local://" + spark_exec_path,
                executor_path=_sys.executable,
            ).to_flyte_idl()

        super(SdkNotebookSparkTask, self).__init__(
            notebook_path,
            inputs,
            outputs,
            _constants.SdkTaskType.SPARK_TASK,
            discovery_version,
            retries,
            deprecated,
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            discoverable,
            timeout,
            environment,
            _json_format.MessageToDict(spark_job),
        )

    def _get_container_definition(self, environment=None, **kwargs):
        """
        :rtype: flytekit.models.task.Container
        """

        return _spark_task.SdkRunnableSparkContainer(
            command=[],
            args=[],
            resources=_task_models.Resources(limits=[], requests=[]),
            env=environment or {},
            config={},
        )



import json
from collections import OrderedDict
from typing import List
from unittest.mock import MagicMock

from kubernetes.client import ApiClient
from kubernetes.client.models import V1Container, V1EnvVar, V1PodSpec, V1ResourceRequirements, V1VolumeMount

from flytekit import Resources, dynamic, task
from flytekit.common.translator import get_serializable
from flytekit.core import context_manager
from flytekit.extend import ExecutionState, Image, ImageConfig, SerializationSettings
from plugins.pod.flytekitplugins.pod.task import Pod, PodFunctionTask


def get_pod_spec():
    a_container = V1Container(
        name="a container",
    )
    a_container.command = ["fee", "fi", "fo", "fum"]
    a_container.volume_mounts = [
        V1VolumeMount(
            name="volume mount",
            mount_path="some/where",
        )
    ]

    pod_spec = V1PodSpec(restart_policy="OnFailure", containers=[a_container, V1Container(name="another container")])
    return pod_spec


def test_pod_task_deserialization():
    pod = Pod(pod_spec=get_pod_spec(), primary_container_name="a container")

    @task(task_config=pod, requests=Resources(cpu="10"), limits=Resources(gpu="2"), environment={"FOO": "bar"})
    def simple_pod_task(i: int):
        pass

    assert isinstance(simple_pod_task, PodFunctionTask)
    assert simple_pod_task.task_config == pod

    default_img = Image(name="default", fqn="test", tag="tag")

    custom = simple_pod_task.get_custom(
        SerializationSettings(
            project="project",
            domain="domain",
            version="version",
            env={"FOO": "baz"},
            image_config=ImageConfig(default_image=default_img, images=[default_img]),
        )
    )

    # Test that custom is correctly serialized by deserializing it with the python API client
    response = MagicMock()
    response.data = json.dumps(custom)
    deserialized_pod_spec = ApiClient().deserialize(response, V1PodSpec)

    assert deserialized_pod_spec.restart_policy == "OnFailure"
    assert len(deserialized_pod_spec.containers) == 2
    primary_container = deserialized_pod_spec.containers[0]
    assert primary_container.name == "a container"
    assert primary_container.args == [
        "pyflyte-execute",
        "--inputs",
        "{{.input}}",
        "--output-prefix",
        "{{.outputPrefix}}",
        "--raw-output-data-prefix",
        "{{.rawOutputDataPrefix}}",
        "--resolver",
        "flytekit.core.python_auto_container.default_task_resolver",
        "--",
        "task-module",
        "plugins.tests.pod.test_pod",
        "task-name",
        "simple_pod_task",
    ]
    assert primary_container.volume_mounts[0].mount_path == "some/where"
    assert primary_container.volume_mounts[0].name == "volume mount"
    assert primary_container.resources == V1ResourceRequirements(limits={"gpu": "2"}, requests={"cpu": "10"})
    assert primary_container.env == [V1EnvVar(name="FOO", value="bar")]
    assert deserialized_pod_spec.containers[1].name == "another container"

    config = simple_pod_task.get_config(
        SerializationSettings(
            project="project",
            domain="domain",
            version="version",
            env={"FOO": "baz"},
            image_config=ImageConfig(default_image=default_img, images=[default_img]),
        )
    )
    assert config["primary_container_name"] == "a container"


def test_pod_task():
    pod = Pod(pod_spec=get_pod_spec(), primary_container_name="a container")

    @task(task_config=pod, requests=Resources(cpu="10"), limits=Resources(gpu="2"), environment={"FOO": "bar"})
    def simple_pod_task(i: int):
        pass

    assert isinstance(simple_pod_task, PodFunctionTask)
    assert simple_pod_task.task_config == pod

    default_img = Image(name="default", fqn="test", tag="tag")

    custom = simple_pod_task.get_custom(
        SerializationSettings(
            project="project",
            domain="domain",
            version="version",
            env={"FOO": "baz"},
            image_config=ImageConfig(default_image=default_img, images=[default_img]),
        )
    )
    assert custom["restartPolicy"] == "OnFailure"
    assert len(custom["containers"]) == 2
    primary_container = custom["containers"][0]
    assert primary_container["name"] == "a container"
    assert primary_container["args"] == [
        "pyflyte-execute",
        "--inputs",
        "{{.input}}",
        "--output-prefix",
        "{{.outputPrefix}}",
        "--raw-output-data-prefix",
        "{{.rawOutputDataPrefix}}",
        "--resolver",
        "flytekit.core.python_auto_container.default_task_resolver",
        "--",
        "task-module",
        "plugins.tests.pod.test_pod",
        "task-name",
        "simple_pod_task",
    ]
    assert primary_container["volumeMounts"][0]["mountPath"] == "some/where"
    assert primary_container["volumeMounts"][0]["name"] == "volume mount"
    assert primary_container["resources"] == {
        "requests": {"cpu": "10"},
        "limits": {"gpu": "2"},
    }
    assert primary_container["env"] == [{"name": "FOO", "value": "bar"}]
    assert custom["containers"][1]["name"] == "another container"


def test_dynamic_pod_task():
    dynamic_pod = Pod(pod_spec=get_pod_spec(), primary_container_name="a container")

    @task
    def t1(a: int) -> int:
        return a + 10

    @dynamic(
        task_config=dynamic_pod, requests=Resources(cpu="10"), limits=Resources(gpu="2"), environment={"FOO": "bar"}
    )
    def dynamic_pod_task(a: int) -> List[int]:
        s = []
        for i in range(a):
            s.append(t1(a=i))
        return s

    assert isinstance(dynamic_pod_task, PodFunctionTask)
    default_img = Image(name="default", fqn="test", tag="tag")

    custom = dynamic_pod_task.get_custom(
        SerializationSettings(
            project="project",
            domain="domain",
            version="version",
            env={"FOO": "baz"},
            image_config=ImageConfig(default_image=default_img, images=[default_img]),
        )
    )
    assert len(custom["containers"]) == 2
    primary_container = custom["containers"][0]
    assert isinstance(dynamic_pod_task.task_config, Pod)
    assert primary_container["resources"] == {
        "requests": {"cpu": "10"},
        "limits": {"gpu": "2"},
    }

    config = dynamic_pod_task.get_config(
        SerializationSettings(
            project="project",
            domain="domain",
            version="version",
            env={"FOO": "baz"},
            image_config=ImageConfig(default_image=default_img, images=[default_img]),
        )
    )
    assert config["primary_container_name"] == "a container"

    with context_manager.FlyteContext.current_context().new_serialization_settings(
        serialization_settings=SerializationSettings(
            project="test_proj",
            domain="test_domain",
            version="abc",
            image_config=ImageConfig(Image(name="name", fqn="image", tag="name")),
            env={},
        )
    ) as ctx:
        with ctx.new_execution_context(mode=ExecutionState.Mode.TASK_EXECUTION) as ctx:
            dynamic_job_spec = dynamic_pod_task.compile_into_workflow(ctx, False, dynamic_pod_task._task_function, a=5)
            assert len(dynamic_job_spec._nodes) == 5


def test_pod_task_undefined_primary():
    pod = Pod(pod_spec=get_pod_spec(), primary_container_name="an undefined container")

    @task(task_config=pod, requests=Resources(cpu="10"), limits=Resources(gpu="2"), environment={"FOO": "bar"})
    def simple_pod_task(i: int):
        pass

    assert isinstance(simple_pod_task, PodFunctionTask)
    assert simple_pod_task.task_config == pod

    default_img = Image(name="default", fqn="test", tag="tag")
    custom = simple_pod_task.get_custom(
        SerializationSettings(
            project="project",
            domain="domain",
            version="version",
            env={"FOO": "baz"},
            image_config=ImageConfig(default_image=default_img, images=[default_img]),
        )
    )

    assert len(custom["containers"]) == 3

    primary_container = custom["containers"][2]
    assert primary_container["name"] == "an undefined container"

    config = simple_pod_task.get_config(
        SerializationSettings(
            project="project",
            domain="domain",
            version="version",
            env={"FOO": "baz"},
            image_config=ImageConfig(default_image=default_img, images=[default_img]),
        )
    )
    assert config["primary_container_name"] == "an undefined container"


def test_pod_task_serialized():
    pod = Pod(pod_spec=get_pod_spec(), primary_container_name="an undefined container")

    @task(task_config=pod, requests=Resources(cpu="10"), limits=Resources(gpu="2"), environment={"FOO": "bar"})
    def simple_pod_task(i: int):
        pass

    assert isinstance(simple_pod_task, PodFunctionTask)
    assert simple_pod_task.task_config == pod

    default_img = Image(name="default", fqn="test", tag="tag")
    ssettings = SerializationSettings(
        project="project",
        domain="domain",
        version="version",
        env={"FOO": "baz"},
        image_config=ImageConfig(default_image=default_img, images=[default_img]),
    )
    serialized = get_serializable(OrderedDict(), ssettings, simple_pod_task)
    assert serialized.task_type_version == 1
    assert serialized.config["primary_container_name"] == "an undefined container"



import os as _os
import subprocess as _subprocess
import tarfile as _tarfile
import tempfile as _tempfile
from pathlib import Path as _Path

import dirhash as _dirhash

from flytekit.interfaces.data import data_proxy as _data_proxy
from flytekit.interfaces.data.data_proxy import Data as _Data

_tmp_versions_dir = "tmp/versions"


def compute_digest(source_dir: _os.PathLike) -> str:
    """
    Walks the entirety of the source dir to compute a deterministic hex digest of the dir contents.
    :param _os.PathLike source_dir:
    :return Text:
    """
    return f"fast{_dirhash.dirhash(source_dir, 'md5', match=['*.py'])}"


def _write_marker(marker: _os.PathLike):
    try:
        open(marker, "x")
    except FileExistsError:
        pass


def filter_tar_file_fn(tarinfo: _tarfile.TarInfo) -> _tarfile.TarInfo:
    """
    Excludes designated file types from tar archive
    :param _tarfile.TarInfo tarinfo:
    :return _tarfile.TarInfo:
    """
    if tarinfo.name.endswith(".pyc"):
        return None
    if tarinfo.name.startswith(".cache"):
        return None
    if "__pycache__" in tarinfo.name:
        return None
    return tarinfo


def get_additional_distribution_loc(remote_location: str, identifier: str) -> str:
    """
    :param Text remote_location:
    :param Text identifier:
    :return Text:
    """
    return _os.path.join(remote_location, "{}.{}".format(identifier, "tar.gz"))


def upload_package(source_dir: _os.PathLike, identifier: str, remote_location: str, dry_run=False) -> str:
    """
    Uploads the contents of the source dir as a tar package to a destination specified by the unique identifier and
    remote_location.
    :param _os.PathLike source_dir:
    :param Text identifier:
    :param Text remote_location:
    :param bool dry_run:
    :return Text:
    """
    tmp_versions_dir = _os.path.join(_os.getcwd(), _tmp_versions_dir)
    _os.makedirs(tmp_versions_dir, exist_ok=True)
    marker = _Path(_os.path.join(tmp_versions_dir, identifier))
    full_remote_path = get_additional_distribution_loc(remote_location, identifier)
    if _os.path.exists(marker):
        print("Local marker for identifier {} already exists, skipping upload".format(identifier))
        return full_remote_path

    if _Data.data_exists(full_remote_path):
        print("Remote file {} already exists, skipping upload".format(full_remote_path))
        _write_marker(marker)
        return full_remote_path

    with _tempfile.NamedTemporaryFile() as fp:
        # Write using gzip
        with _tarfile.open(fp.name, "w:gz") as tar:
            tar.add(source_dir, arcname="", filter=filter_tar_file_fn)
        if dry_run:
            print("Would upload {} to {}".format(fp.name, full_remote_path))
        else:
            _Data.put_data(fp.name, full_remote_path)
            print("Uploaded {} to {}".format(fp.name, full_remote_path))

    # Finally, touch the marker file so we have a flag in the future to avoid re-uploading the package dir as an
    # optimization
    _write_marker(marker)
    return full_remote_path


def download_distribution(additional_distribution: str, destination: str):
    """
    Downloads a remote code distribution and overwrites any local files.
    :param Text additional_distribution:
    :param _os.PathLike destination:
    """
    _data_proxy.Data.get_data(additional_distribution, destination)
    tarfile_name = _os.path.basename(additional_distribution)
    file_suffix = _Path(tarfile_name).suffixes
    if len(file_suffix) != 2 or file_suffix[0] != ".tar" or file_suffix[1] != ".gz":
        raise ValueError("Unrecognized additional distribution format for {}".format(additional_distribution))

    # This will overwrite the existing user flyte workflow code in the current working code dir.
    result = _subprocess.run(
        ["tar", "-xvf", _os.path.join(destination, tarfile_name), "-C", destination],
        stdout=_subprocess.PIPE,
    )
    result.check_returncode()



import abc as _abc

from flyteidl.core.literals_pb2 import Literal

from flytekit.common import sdk_bases as _sdk_bases
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.models import common as _common_models
from flytekit.models import literals as _literal_models


class FlyteSdkType(_sdk_bases.ExtendedSdkType, metaclass=_common_models.FlyteABCMeta):
    @_abc.abstractmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        pass

    @_abc.abstractmethod
    def from_python_std(cls, t_value):
        """
        :param T t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        pass

    @_abc.abstractmethod
    def from_string(cls, string_value):
        """
        :param Text string_value: It is up to each individual object to implement this.
        :rtype: FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        pass

    @_abc.abstractmethod
    def promote_from_model(cls, literal):
        """
        :param flytekit.models.literals.Literal literal:
        :rtype: FlyteSdkValue
        """
        pass

    @_abc.abstractmethod
    def to_flyte_literal_type(cls):
        """
        :rtype: flytekit.models.types.LiteralType
        """
        pass

    def __hash__(cls):
        return hash(cls.to_flyte_literal_type())


class FlyteSdkValue(_literal_models.Literal, metaclass=FlyteSdkType):
    @classmethod
    def from_flyte_idl(cls, pb2_object: Literal):
        """
        :param flyteidl.core.literals_pb2.Literal pb2_object:
        :rtype: FlyteSdkValue
        """
        literal = _literal_models.Literal.from_flyte_idl(pb2_object)
        if literal.scalar is not None and literal.scalar.none_type is not None:
            return Void()
        return cls.promote_from_model(literal)

    @_abc.abstractmethod
    def to_python_std(self):
        pass


class InstantiableType(FlyteSdkType, metaclass=_common_models.FlyteABCMeta):
    @_abc.abstractmethod
    def __call__(cls, *args, **kwargs):
        """
        TODO: Figure out generics for type hinting.

        :rtype: T
        """
        return super(InstantiableType, cls).__call__(*args, **kwargs)


class Void(FlyteSdkValue):
    @classmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        return True

    @classmethod
    def from_python_std(cls, t_value):
        """
        :param T t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        return cls()

    @classmethod
    def to_flyte_literal_type(cls):
        """
        :rtype: flytekit.models.types.LiteralType
        """
        raise _user_exceptions.FlyteAssertion(
            "A Void type does not have a literal type and cannot be used in this " "manner."
        )

    @classmethod
    def promote_from_model(cls, _):
        """
        Creates an object of this type from the model primitive defining it.
        :param flytekit.models.literals.Literal _:
        :rtype: Void
        """
        return cls()

    @classmethod
    def short_class_string(cls):
        """
        :rtype: Text
        """
        return "Void"

    def __init__(self):
        super(Void, self).__init__(scalar=_literal_models.Scalar(none_type=_literal_models.Void()))

    def to_python_std(self):
        """
        :rtype: NoneType
        """
        return None

    def short_string(self):
        """
        :rtype: Text
        """
        return "Void()"



import contextlib
import importlib
import os
import pkgutil
import sys
from typing import Any, Iterator, List, Union

from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.local_workflow import SdkRunnableWorkflow as _SdkRunnableWorkflow
from flytekit.common.mixins import registerable as _registerable


def iterate_modules(pkgs):
    for package_name in pkgs:
        package = importlib.import_module(package_name)
        yield package

        # Check if package is a python file. If so, there is no reason to walk.
        if not hasattr(package, "__path__"):
            continue

        for _, name, _ in pkgutil.walk_packages(package.__path__, prefix="{}.".format(package_name)):
            yield importlib.import_module(name)


@contextlib.contextmanager
def add_sys_path(path: Union[str, os.PathLike]) -> Iterator[None]:
    """Temporarily add given path to `sys.path`."""
    path = os.fspath(path)
    try:
        sys.path.insert(0, path)
        yield
    finally:
        sys.path.remove(path)


def just_load_modules(pkgs: List[str]):
    """
    This one differs from the above in that we don't yield anything, just load all the modules.
    """
    for package_name in pkgs:
        package = importlib.import_module(package_name)
        for _, name, _ in pkgutil.walk_packages(package.__path__, prefix="{}.".format(package_name)):
            importlib.import_module(name)


def load_workflow_modules(pkgs):
    """
    Load all modules and packages at and under the given package.  Used for finding workflows/tasks to register.

    :param list[Text] pkgs: List of dot separated string containing paths folders (packages) containing
        the modules (python files)
    :raises ImportError
    """
    for _ in iterate_modules(pkgs):
        pass


def _topo_sort_helper(
    obj,
    entity_to_module_key,
    visited,
    recursion_set,
    recursion_stack,
    include_entities,
    ignore_entities,
    detect_unreferenced_entities,
):
    visited.add(obj)
    recursion_stack.append(obj)
    if obj in recursion_set:
        raise _user_exceptions.FlyteAssertion(
            "A cyclical dependency was detected during topological sort of entities.  "
            "Cycle path was:\n\n\t{}".format("\n\t".join(p for p in recursion_stack[recursion_set[obj] :]))
        )
    recursion_set[obj] = len(recursion_stack) - 1

    if isinstance(obj, _registerable.HasDependencies):
        for upstream in obj.upstream_entities:
            if upstream.has_registered:
                continue
            if upstream not in visited:
                for m1, k1, o1 in _topo_sort_helper(
                    upstream,
                    entity_to_module_key,
                    visited,
                    recursion_set,
                    recursion_stack,
                    include_entities,
                    ignore_entities,
                    detect_unreferenced_entities,
                ):
                    if not o1.has_registered:
                        yield m1, k1, o1

    recursion_stack.pop()
    del recursion_set[obj]

    if isinstance(obj, include_entities) or not isinstance(obj, ignore_entities):
        if obj in entity_to_module_key:
            yield entity_to_module_key[obj] + (obj,)
        elif detect_unreferenced_entities:
            raise _user_exceptions.FlyteAssertion(
                f"An entity ({obj.id}) was not found in modules accessible from the workflow packages configuration.  Please "
                f"ensure that entities in '{obj.instantiated_in}' are moved to a configured packaged, or adjust the configuration."
            )


def _get_entity_to_module(pkgs):
    entity_to_module_key = {}
    for m in iterate_modules(pkgs):
        for k in dir(m):
            o = m.__dict__[k]
            if isinstance(o, _registerable.RegisterableEntity) and not o.has_registered:
                if o.instantiated_in == m.__name__:
                    entity_to_module_key[o] = (m, k)
                    if isinstance(o, _SdkRunnableWorkflow) and o.should_create_default_launch_plan:
                        # SDK should create a default launch plan for a workflow.  This is a special-case to simplify
                        # authoring of workflows.
                        entity_to_module_key[o.create_launch_plan()] = (m, k)
    return entity_to_module_key


def load_module_object_for_type(pkgs, t, additional_path=None):
    def iterate():
        entity_to_module_key = {}
        for m in iterate_modules(pkgs):
            for k in dir(m):
                o = m.__dict__[k]
                if isinstance(o, t):
                    entity_to_module_key[o] = (m.__name__, k)
        return entity_to_module_key

    if additional_path is not None:
        with add_sys_path(additional_path):
            return iterate()
    else:
        return iterate()


def iterate_registerable_entities_in_order(
    pkgs,
    local_source_root=None,
    ignore_entities=None,
    include_entities=None,
    detect_unreferenced_entities=True,
):
    """
    This function will iterate all discovered entities in the given package list.  It will then attempt to
    topologically sort such that any entity with a dependency on another comes later in the list.  Note that workflows
    can reference other workflows and launch plans.
    :param list[Text] pkgs:
    :param Text local_source_root:
    :param set[type] ignore_entities: If specified, ignore these entities while doing a topological sort.  All other
        entities will be taken.  Only one of ignore_entities or include_entities can be set.
    :param set[type] include_entities: If specified, include these entities while doing a topological sort.  All
        other entities will be ignored.  Only one of ignore_entities or include_entities can be set.
    :param bool detect_unreferenced_entities: If true, we will raise exceptions on entities not included in the package
        configuration.
    :rtype: module, Text, flytekit.common.mixins.registerable.RegisterableEntity
    """
    if ignore_entities and include_entities:
        raise _user_exceptions.FlyteAssertion("ignore_entities and include_entities cannot both be set")
    elif not ignore_entities and not include_entities:
        include_entities = (object,)
        ignore_entities = tuple()
    else:
        ignore_entities = tuple(list(ignore_entities or set([object])))
        include_entities = tuple(list(include_entities or set()))

    if local_source_root is not None:
        with add_sys_path(local_source_root):
            entity_to_module_key = _get_entity_to_module(pkgs)
    else:
        entity_to_module_key = _get_entity_to_module(pkgs)

    visited = set()
    for o in entity_to_module_key.keys():
        if o not in visited:
            recursion_set = dict()
            recursion_stack = []
            for m, k, o2 in _topo_sort_helper(
                o,
                entity_to_module_key,
                visited,
                recursion_set,
                recursion_stack,
                include_entities,
                ignore_entities,
                detect_unreferenced_entities=detect_unreferenced_entities,
            ):
                yield m, k, o2


def load_object_from_module(object_location: str) -> Any:
    """
    # TODO: Handle corner cases, like where the first part is [] maybe
    """
    class_obj = object_location.split(".")
    class_obj_mod = class_obj[:-1]  # e.g. ['flytekit', 'core', 'python_auto_container']
    class_obj_key = class_obj[-1]  # e.g. 'default_task_class_obj'
    class_obj_mod = importlib.import_module(".".join(class_obj_mod))
    return getattr(class_obj_mod, class_obj_key)



import importlib as _importer
from typing import Type

from flytekit.common.exceptions import system as _system_exceptions
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.types import base_sdk_types as _base_sdk_types
from flytekit.common.types import blobs as _blobs
from flytekit.common.types import containers as _container_types
from flytekit.common.types import helpers as _helpers
from flytekit.common.types import primitives as _primitive_types
from flytekit.common.types import proto as _proto
from flytekit.common.types import schema as _schema
from flytekit.models import types as _literal_type_models
from flytekit.models.core import types as _core_types


def _load_type_from_tag(tag: str) -> Type:
    """
    Loads python type from tag
    """

    if "." not in tag:
        raise _user_exceptions.FlyteValueException(
            tag,
            "Protobuf tag must include at least one '.' to delineate package and object name.",
        )

    module, name = tag.rsplit(".", 1)
    try:
        pb_module = _importer.import_module(module)
    except ImportError:
        raise _user_exceptions.FlyteAssertion(
            "Could not resolve the protobuf definition @ {}.  Is the protobuf library installed?".format(module)
        )

    if not hasattr(pb_module, name):
        raise _user_exceptions.FlyteAssertion("Could not find the protobuf named: {} @ {}.".format(name, module))

    return getattr(pb_module, name)


def _proto_sdk_type_from_tag(tag):
    """
    :param Text tag:
    :rtype: _proto.Protobuf
    """
    return _proto.create_protobuf(_load_type_from_tag(tag))


def _generic_proto_sdk_type_from_tag(tag: str) -> Type[_proto.GenericProtobuf]:
    """
    :param Text tag:
    :rtype: _proto.GenericProtobuf
    """

    return _proto.create_generic(_load_type_from_tag(tag))


class FlyteDefaultTypeEngine(object):
    _SIMPLE_TYPE_LOOKUP_TABLE = {
        _literal_type_models.SimpleType.INTEGER: _primitive_types.Integer,
        _literal_type_models.SimpleType.FLOAT: _primitive_types.Float,
        _literal_type_models.SimpleType.BOOLEAN: _primitive_types.Boolean,
        _literal_type_models.SimpleType.DATETIME: _primitive_types.Datetime,
        _literal_type_models.SimpleType.DURATION: _primitive_types.Timedelta,
        _literal_type_models.SimpleType.NONE: _base_sdk_types.Void,
        _literal_type_models.SimpleType.STRING: _primitive_types.String,
        _literal_type_models.SimpleType.STRUCT: _primitive_types.Generic,
    }

    def python_std_to_sdk_type(self, t):
        """
        :param T t: User input.  Should be of the form: Types.Integer, [Types.Integer], {Types.String:
        Types.Integer}, etc.
        :rtype: flytekit.common.types.base_sdk_types.FlyteSdkType
        """
        if isinstance(t, list):
            if len(t) != 1:
                raise _user_exceptions.FlyteAssertion(
                    "When specifying a list type, there must be exactly one element in "
                    "the list describing the contained type."
                )
            return _container_types.List(_helpers.python_std_to_sdk_type(t[0]))
        elif isinstance(t, dict):
            raise _user_exceptions.FlyteAssertion("Map types are not yet implemented.")
        elif isinstance(t, _base_sdk_types.FlyteSdkType):
            return t
        else:
            raise _user_exceptions.FlyteTypeException(
                type(t),
                _base_sdk_types.FlyteSdkType,
                additional_msg="Should be of form similar to: Types.Integer, [Types.Integer], {Types.String: "
                "Types.Integer}",
                received_value=t,
            )

    def get_sdk_type_from_literal_type(self, literal_type):
        """
        :param flytekit.models.types.LiteralType literal_type:
        :rtype: flytekit.common.types.base_sdk_types.FlyteSdkType
        """
        if literal_type.collection_type is not None:
            return _container_types.List(_helpers.get_sdk_type_from_literal_type(literal_type.collection_type))
        elif literal_type.map_value_type is not None:
            raise NotImplementedError("TODO: Implement map")
        elif literal_type.schema is not None:
            return _schema.schema_instantiator_from_proto(literal_type.schema)
        elif literal_type.blob is not None:
            return self._get_blob_impl_from_type(literal_type.blob)
        elif literal_type.simple is not None:
            if (
                literal_type.simple == _literal_type_models.SimpleType.BINARY
                and _proto.Protobuf.PB_FIELD_KEY in literal_type.metadata
            ):
                return _proto_sdk_type_from_tag(literal_type.metadata[_proto.Protobuf.PB_FIELD_KEY])
            if (
                literal_type.simple == _literal_type_models.SimpleType.STRUCT
                and literal_type.metadata
                and _proto.Protobuf.PB_FIELD_KEY in literal_type.metadata
            ):
                return _generic_proto_sdk_type_from_tag(literal_type.metadata[_proto.Protobuf.PB_FIELD_KEY])
            sdk_type = self._SIMPLE_TYPE_LOOKUP_TABLE.get(literal_type.simple)
            if sdk_type is None:
                raise NotImplementedError(
                    "We haven't implemented this type yet:  Simple type={}".format(literal_type.simple)
                )
            return sdk_type
        else:
            raise _system_exceptions.FlyteSystemAssertion(
                "An unrecognized literal type was received: {}".format(literal_type)
            )

    def infer_sdk_type_from_literal(self, literal):  # noqa
        """
        :param flytekit.models.literals.Literal literal:
        :rtype: flytekit.common.types.base_sdk_types.FlyteSdkType
        """
        if literal.collection is not None:
            if len(literal.collection.literals) > 0:
                sdk_type = _container_types.List(_helpers.infer_sdk_type_from_literal(literal.collection.literals[0]))
            else:
                sdk_type = _container_types.List(_base_sdk_types.Void)
        elif literal.map is not None:
            raise NotImplementedError("TODO: Implement map")
        elif literal.scalar.blob is not None:
            sdk_type = self._get_blob_impl_from_type(literal.scalar.blob.metadata.type)
        elif literal.scalar.none_type is not None:
            sdk_type = _base_sdk_types.Void
        elif literal.scalar.schema is not None:
            sdk_type = _schema.schema_instantiator_from_proto(literal.scalar.schema.type)
        elif literal.scalar.error is not None:
            raise NotImplementedError("TODO: Implement error from literal map")
        elif literal.scalar.generic is not None:
            sdk_type = _primitive_types.Generic
        elif literal.scalar.binary is not None:
            if literal.scalar.binary.tag.startswith(_proto.Protobuf.TAG_PREFIX):
                sdk_type = _proto_sdk_type_from_tag(literal.scalar.binary.tag[len(_proto.Protobuf.TAG_PREFIX) :])
            else:
                raise NotImplementedError("TODO: Binary is only supported for protobuf types currently")
        elif literal.scalar.primitive.boolean is not None:
            sdk_type = _primitive_types.Boolean
        elif literal.scalar.primitive.datetime is not None:
            sdk_type = _primitive_types.Datetime
        elif literal.scalar.primitive.duration is not None:
            sdk_type = _primitive_types.Timedelta
        elif literal.scalar.primitive.float_value is not None:
            sdk_type = _primitive_types.Float
        elif literal.scalar.primitive.integer is not None:
            sdk_type = _primitive_types.Integer
        elif literal.scalar.primitive.string_value is not None:
            sdk_type = _primitive_types.String
        else:
            raise _system_exceptions.FlyteSystemAssertion("Received unknown literal: {}".format(literal))
        return sdk_type

    def _get_blob_impl_from_type(self, blob_type):
        """
        :param flytekit.models.core.types.BlobType blob_type:
        :rtype: flytekit.common.types.base_sdk_types.FlyteSdkType
        """
        if blob_type.dimensionality == _core_types.BlobType.BlobDimensionality.SINGLE:
            if blob_type.format == "csv":
                return _blobs.CSV
            else:
                return _blobs.Blob
        elif blob_type.dimensionality == _core_types.BlobType.BlobDimensionality.MULTIPART:
            if blob_type.format == "csv":
                return _blobs.MultiPartCSV
            else:
                return _blobs.MultiPartBlob
        raise _system_exceptions.FlyteSystemAssertion(
            "Flyte's base type engine does not support this type of blob. Value: {}".format(blob_type)
        )



import datetime as _datetime

import six as _six
from google.protobuf.json_format import MessageToDict as _MessageToDict

from flytekit import __version__
from flytekit.common import constants as _constants
from flytekit.common import interface as _interface
from flytekit.common.exceptions import scopes as _exception_scopes
from flytekit.common.tasks import task as _base_task
from flytekit.common.types import helpers as _type_helpers
from flytekit.models import interface as _interface_model
from flytekit.models import literals as _literals
from flytekit.models import presto as _presto_models
from flytekit.models import task as _task_model
from flytekit.models import types as _types


class SdkPrestoTask(_base_task.SdkTask):
    """
    This class includes the logic for building a task that executes as a Presto task.
    """

    def __init__(
        self,
        statement,
        output_schema,
        routing_group=None,
        catalog=None,
        schema=None,
        task_inputs=None,
        interruptible=False,
        discoverable=False,
        discovery_version=None,
        retries=1,
        timeout=None,
        deprecated=None,
    ):
        """
        :param Text statement: Presto query specification
        :param flytekit.common.types.schema.Schema output_schema: Schema that represents that data queried from Presto
        :param Text routing_group: The routing group that a Presto query should be sent to for the given environment
        :param Text catalog: The catalog to set for the given Presto query
        :param Text schema: The schema to set for the given Presto query
        :param dict[Text,flytekit.common.types.base_sdk_types.FlyteSdkType] task_inputs: Optional inputs to the Presto task
        :param bool discoverable:
        :param Text discovery_version: String describing the version for task discovery purposes
        :param int retries: Number of retries to attempt
        :param datetime.timedelta timeout:
        :param Text deprecated: This string can be used to mark the task as deprecated.  Consumers of the task will
            receive deprecation warnings.
        """

        # Set as class fields which are used down below to configure implicit
        # parameters
        self._routing_group = routing_group or ""
        self._catalog = catalog or ""
        self._schema = schema or ""

        metadata = _task_model.TaskMetadata(
            discoverable,
            # This needs to have the proper version reflected in it
            _task_model.RuntimeMetadata(_task_model.RuntimeMetadata.RuntimeType.FLYTE_SDK, __version__, "python"),
            timeout or _datetime.timedelta(seconds=0),
            _literals.RetryStrategy(retries),
            interruptible,
            discovery_version,
            deprecated,
        )

        presto_query = _presto_models.PrestoQuery(
            routing_group=routing_group or "",
            catalog=catalog or "",
            schema=schema or "",
            statement=statement,
        )

        # Here we set the routing_group, catalog, and schema as implicit
        # parameters for caching purposes
        i = _interface.TypedInterface(
            {
                "__implicit_routing_group": _interface_model.Variable(
                    type=_types.LiteralType(simple=_types.SimpleType.STRING),
                    description="The routing group set as an implicit input",
                ),
                "__implicit_catalog": _interface_model.Variable(
                    type=_types.LiteralType(simple=_types.SimpleType.STRING),
                    description="The catalog set as an implicit input",
                ),
                "__implicit_schema": _interface_model.Variable(
                    type=_types.LiteralType(simple=_types.SimpleType.STRING),
                    description="The schema set as an implicit input",
                ),
            },
            {
                # Set the schema for the Presto query as an output
                "results": _interface_model.Variable(
                    type=_types.LiteralType(schema=output_schema.schema_type),
                    description="The schema for the Presto query",
                )
            },
        )

        super(SdkPrestoTask, self).__init__(
            _constants.SdkTaskType.PRESTO_TASK,
            metadata,
            i,
            _MessageToDict(presto_query.to_flyte_idl()),
        )

        # Set user provided inputs
        task_inputs(self)

    def _add_implicit_inputs(self, inputs):
        """
        :param dict[Text,Any] inputs:
        :param inputs:
        :return:
        """
        inputs["__implicit_routing_group"] = self.routing_group
        inputs["__implicit_catalog"] = self.catalog
        inputs["__implicit_schema"] = self.schema
        return inputs

    # Override method in order to set the implicit inputs
    def __call__(self, *args, **kwargs):
        kwargs = self._add_implicit_inputs(kwargs)

        return super(SdkPrestoTask, self).__call__(*args, **kwargs)

    # Override method in order to set the implicit inputs
    def _python_std_input_map_to_literal_map(self, inputs):
        """
        :param dict[Text,Any] inputs: A dictionary of Python standard inputs that will be type-checked and compiled
            to a LiteralMap
        :rtype: flytekit.models.literals.LiteralMap
        """
        inputs = self._add_implicit_inputs(inputs)
        return _type_helpers.pack_python_std_map_to_literal_map(
            inputs,
            {k: _type_helpers.get_sdk_type_from_literal_type(v.type) for k, v in _six.iteritems(self.interface.inputs)},
        )

    @_exception_scopes.system_entry_point
    def add_inputs(self, inputs):
        """
        Adds the inputs to this task.  This can be called multiple times, but it will fail if an input with a given
        name is added more than once, a name collides with an output, or if the name doesn't exist as an arg name in
        the wrapped function.
        :param dict[Text, flytekit.models.interface.Variable] inputs: names and variables
        """
        self._validate_inputs(inputs)
        self.interface.inputs.update(inputs)

    @property
    def routing_group(self):
        """
        The routing group that a Presto query should be sent to for the given environment
        :rtype: Text
        """
        return self._routing_group

    @property
    def catalog(self):
        """
        The catalog to set for the given Presto query
        :rtype: Text
        """
        return self._catalog

    @property
    def schema(self):
        """
        The schema to set for the given Presto query
        :rtype: Text
        """
        return self._schema



from flytekitplugins.kftensorflow import TfJob

from flytekit import Resources, task
from flytekit.extend import Image, ImageConfig, SerializationSettings


def test_tensorflow_task():
    @task(
        task_config=TfJob(
            num_workers=10, per_replica_requests=Resources(cpu="1"), num_ps_replicas=1, num_chief_replicas=1
        ),
        cache=True,
        cache_version="1",
    )
    def my_tensorflow_task(x: int, y: str) -> int:
        return x

    assert my_tensorflow_task(x=10, y="hello") == 10

    assert my_tensorflow_task.task_config is not None

    default_img = Image(name="default", fqn="test", tag="tag")
    settings = SerializationSettings(
        project="project",
        domain="domain",
        version="version",
        env={"FOO": "baz"},
        image_config=ImageConfig(default_image=default_img, images=[default_img]),
    )

    assert my_tensorflow_task.get_custom(settings) == {"workers": 10, "psReplicas": 1, "chiefReplicas": 1}
    assert my_tensorflow_task.resources.limits == Resources()
    assert my_tensorflow_task.resources.requests == Resources(cpu="1")
    assert my_tensorflow_task.task_type == "tensorflow"



import importlib as _importlib

from flytekit.common.exceptions import scopes as _exception_scopes
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.configuration import sdk as _sdk_config

_ENGINE_NAME_TO_MODULES_CACHE = {
    "flyte": ("flytekit.engines.flyte.engine", "FlyteEngineFactory", None),
    "unit": ("flytekit.engines.unit.engine", "UnitTestEngineFactory", None),
    # 'local': ('flytekit.engines.local.engine', 'EngineObjectFactory', None)
}


def get_engine(engine_name=None):
    """
    :param Text engine_name:
    :rtype: flytekit.engines.common.BaseExecutionEngineFactory
    """
    engine_name = engine_name or _sdk_config.EXECUTION_ENGINE.get()

    # TODO: Allow users to plug-in their own engine code via a config
    if engine_name not in _ENGINE_NAME_TO_MODULES_CACHE:
        raise _user_exceptions.FlyteValueException(
            engine_name,
            "Could not load an engine with the identifier '{}'.  Known engines are: {}".format(
                engine_name, list(_ENGINE_NAME_TO_MODULES_CACHE.keys())
            ),
        )

    module_path, attr, engine_impl = _ENGINE_NAME_TO_MODULES_CACHE[engine_name]
    if engine_impl is None:
        module = _exception_scopes.user_entry_point(_importlib.import_module)(module_path)

        if not hasattr(module, attr):
            raise _user_exceptions.FlyteValueException(
                module,
                "Failed to load the engine because the attribute named '{}' could not be found"
                "in the module '{}'.".format(attr, module_path),
            )

        engine_impl = getattr(module, attr)()
        _ENGINE_NAME_TO_MODULES_CACHE[engine_name] = (module_path, attr, engine_impl)

    return engine_impl



import uuid as _uuid
from typing import Any, List

from flytekit.common.exceptions import scopes as _exception_scopes
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.control_plane import identifier as _identifier
from flytekit.control_plane import interface as _interface
from flytekit.control_plane import nodes as _nodes
from flytekit.control_plane import workflow_execution as _workflow_execution
from flytekit.engines.flyte import engine as _flyte_engine
from flytekit.models import common as _common_models
from flytekit.models import execution as _execution_models
from flytekit.models import interface as _interface_models
from flytekit.models import launch_plan as _launch_plan_models
from flytekit.models import literals as _literal_models
from flytekit.models.core import identifier as _identifier_model


class FlyteLaunchPlan(_launch_plan_models.LaunchPlanSpec):
    def __init__(self, *args, **kwargs):
        super(FlyteLaunchPlan, self).__init__(*args, **kwargs)
        # Set all the attributes we expect this class to have
        self._id = None

        # The interface is not set explicitly unless fetched in an engine context
        self._interface = None

    @classmethod
    def promote_from_model(cls, model: _launch_plan_models.LaunchPlanSpec) -> "FlyteLaunchPlan":
        return cls(
            workflow_id=_identifier.Identifier.promote_from_model(model.workflow_id),
            default_inputs=_interface_models.ParameterMap(model.default_inputs.parameters),
            fixed_inputs=model.fixed_inputs,
            entity_metadata=model.entity_metadata,
            labels=model.labels,
            annotations=model.annotations,
            auth_role=model.auth_role,
            raw_output_data_config=model.raw_output_data_config,
        )

    @_exception_scopes.system_entry_point
    def register(self, project, domain, name, version):
        # NOTE: does this need to be implemented in the control plane?
        pass

    @classmethod
    @_exception_scopes.system_entry_point
    def fetch(cls, project: str, domain: str, name: str, version: str) -> "FlyteLaunchPlan":
        """
        This function uses the engine loader to call create a hydrated task from Admin.
        :param project:
        :param domain:
        :param name:
        :param version:
        """
        from flytekit.control_plane import workflow as _workflow

        launch_plan_id = _identifier.Identifier(
            _identifier_model.ResourceType.LAUNCH_PLAN, project, domain, name, version
        )

        lp = _flyte_engine.get_client().get_launch_plan(launch_plan_id)
        flyte_lp = cls.promote_from_model(lp.spec)
        flyte_lp._id = lp.id

        # TODO: Add a test for this, and this function as a whole
        wf_id = flyte_lp.workflow_id
        lp_wf = _workflow.FlyteWorkflow.fetch(wf_id.project, wf_id.domain, wf_id.name, wf_id.version)
        flyte_lp._interface = lp_wf.interface
        return flyte_lp

    @_exception_scopes.system_entry_point
    def serialize(self):
        """
        Serializing a launch plan should produce an object similar to what the registration step produces,
        in preparation for actual registration to Admin.

        :rtype: flyteidl.admin.launch_plan_pb2.LaunchPlan
        """
        # NOTE: does this need to be implemented in the control plane?
        pass

    @property
    def id(self) -> _identifier.Identifier:
        return self._id

    @property
    def is_scheduled(self) -> bool:
        if self.entity_metadata.schedule.cron_expression:
            return True
        elif self.entity_metadata.schedule.rate and self.entity_metadata.schedule.rate.value:
            return True
        elif self.entity_metadata.schedule.cron_schedule and self.entity_metadata.schedule.cron_schedule.schedule:
            return True
        else:
            return False

    @property
    def workflow_id(self) -> _identifier.Identifier:
        return self._workflow_id

    @property
    def interface(self) -> _interface.TypedInterface:
        """
        The interface is not technically part of the admin.LaunchPlanSpec in the IDL, however the workflow ID is, and
        from the workflow ID, fetch will fill in the interface. This is nice because then you can __call__ the=
        object and get a node.
        """
        return self._interface

    @property
    def resource_type(self) -> _identifier_model.ResourceType:
        return _identifier_model.ResourceType.LAUNCH_PLAN

    @property
    def entity_type_text(self) -> str:
        return "Launch Plan"

    @_exception_scopes.system_entry_point
    def validate(self):
        # TODO: Validate workflow is satisfied
        pass

    @_exception_scopes.system_entry_point
    def update(self, state: _launch_plan_models.LaunchPlanState):
        if not self.id:
            raise _user_exceptions.FlyteAssertion(
                "Failed to update launch plan because the launch plan's ID is not set. Please call register to fetch "
                "or register the identifier first"
            )
        return _flyte_engine.get_client().update_launch_plan(self.id, state)

    @_exception_scopes.system_entry_point
    def launch_with_literals(
        self,
        project: str,
        domain: str,
        literal_inputs: _literal_models.LiteralMap,
        name: str = None,
        notification_overrides: List[_common_models.Notification] = None,
        label_overrides: _common_models.Labels = None,
        annotation_overrides: _common_models.Annotations = None,
    ) -> _workflow_execution.FlyteWorkflowExecution:
        """
        Executes the launch plan and returns the execution identifier.  This version of execution is meant for when
        you already have a LiteralMap of inputs.

        :param project:
        :param domain:
        :param literal_inputs: Inputs to the execution.
        :param name: If specified, an execution will be created with this name.  Note: the name must
            be unique within the context of the project and domain.
        :param notification_overrides: If specified, these are the notifications that will be honored for this
            execution. An empty list signals to disable all notifications.
        :param label_overrides:
        :param annotation_overrides:
        """
        # Kubernetes requires names starting with an alphabet for some resources.
        name = name or "f" + _uuid.uuid4().hex[:19]
        disable_all = notification_overrides == []
        if disable_all:
            notification_overrides = None
        else:
            notification_overrides = _execution_models.NotificationList(notification_overrides or [])
            disable_all = None

        client = _flyte_engine.get_client()
        try:
            exec_id = client.create_execution(
                project,
                domain,
                name,
                _execution_models.ExecutionSpec(
                    self.id,
                    _execution_models.ExecutionMetadata(
                        _execution_models.ExecutionMetadata.ExecutionMode.MANUAL,
                        "sdk",  # TODO: get principle
                        0,  # TODO: Detect nesting
                    ),
                    notifications=notification_overrides,
                    disable_all=disable_all,
                    labels=label_overrides,
                    annotations=annotation_overrides,
                ),
                literal_inputs,
            )
        except _user_exceptions.FlyteEntityAlreadyExistsException:
            exec_id = _identifier.WorkflowExecutionIdentifier(project, domain, name)
        return _workflow_execution.FlyteWorkflowExecution.promote_from_model(client.get_execution(exec_id))

    @_exception_scopes.system_entry_point
    def __call__(self, *args, **input_map: Any) -> _nodes.FlyteNode:
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"FlyteLaunchPlan(ID: {self.id} Interface: {self.interface} WF ID: {self.workflow_id})"



import abc as _abc


class TypeEngine(object, metaclass=_abc.ABCMeta):
    @_abc.abstractmethod
    def python_std_to_sdk_type(self, t):
        """
        Converts a standard format for specifying types in Python to the Flyte typing structure.
        :param T t: User input.  Usually of the form: Types.Integer, [Types.Integer], {Types.String:
        Types.Integer}, etc.
        :rtype: flytekit.common.types.base_sdk_types.FlyteSdkType
        """
        pass

    @_abc.abstractmethod
    def get_sdk_type_from_literal_type(self, literal_type):
        """
        Takes the Flyte spec language and converts to an SDK object.
        :param flytekit.models.types.LiteralType literal_type:
        :rtype: flytekit.common.types.base_sdk_types.FlyteSdkType
        """
        pass

    @_abc.abstractmethod
    def infer_sdk_type_from_literal(self, literal):
        """
        From a literal value, we infer the correct SDK type.
        :param flytekit.models.literals.Literal literal:
        :rtype: flytekit.common.types.base_sdk_types.FlyteSdkType
        """
        pass



from flytekit.models.sagemaker import hpo_job as _hpo_models
from flytekit.models.sagemaker import parameter_ranges as _parameter_range_models
from flytekit.sdk import types as _sdk_types

HyperparameterTuningJobConfig = _sdk_types.Types.GenericProto(_hpo_models.HyperparameterTuningJobConfig)

ParameterRange = _sdk_types.Types.GenericProto(_parameter_range_models.ParameterRangeOneOf)



from __future__ import annotations

import importlib as _importlib
import sys as _sys
import types as _types
from typing import List


class LazyLoadPlugin(object):
    LAZY_LOADING_PLUGINS = {}

    def __init__(self, plugin_name, plugin_requirements, related_modules: List[_LazyLoadModule]):
        """
        :param Text plugin_name:
        :param list[Text] plugin_requirements:
        :param list[LazyLoadModule] related_modules:
        """
        type(self).LAZY_LOADING_PLUGINS[plugin_name] = plugin_requirements
        for m in related_modules:
            type(m).tag_with_plugin(plugin_name)

    @classmethod
    def get_extras_require(cls):
        """
        :rtype: dict[Text,list[Text]]
        """
        d = cls.LAZY_LOADING_PLUGINS.copy()
        all_plugins_spark2 = []
        all_plugins_spark3 = []
        for k in d:
            # Default to Spark 2.4.x in all-spark2 and Spark 3.x in all-spark3.
            if k != "spark3":
                all_plugins_spark2.extend(d[k])
            if k != "spark":
                all_plugins_spark3.extend(d[k])

        d["all-spark2.4"] = all_plugins_spark2
        d["all-spark3"] = all_plugins_spark3
        # all points to Spark 3.x.
        # Spark 2.4 to be fully removed in a future release.
        d["all"] = all_plugins_spark3
        return d


def lazy_load_module(module: str) -> _types.ModuleType:
    """
    :param Text module:
    :rtype: _types.ModuleType
    """

    class LazyLoadModule(_LazyLoadModule):
        _module = module
        _lazy_submodules = dict()
        _plugins = []

    return LazyLoadModule(module)


class _LazyLoadModule(_types.ModuleType):
    _ERROR_MSG_FMT = (
        "Attempting to use a plugin functionality that requires module "
        "`{module}`, but it couldn't be loaded. Please pip install at least one of {plugins} or "
        "`flytekit[all]` to get these dependencies.\n"
        "\n"
        "Original message: {msg}"
    )

    @classmethod
    def _load(cls):
        module = _sys.modules.get(cls._module)
        if not module:
            try:
                module = _importlib.import_module(cls._module)
            except ImportError as e:
                raise ImportError(cls._ERROR_MSG_FMT.format(module=cls._module, plugins=cls._plugins, msg=e))
        return module

    def __getattribute__(self, item):
        if item in type(self)._lazy_submodules:
            return type(self)._lazy_submodules[item]
        m = type(self)._load()
        return getattr(m, item)

    def __setattr__(self, key, value):
        m = type(self)._load()
        return setattr(m, key, value)

    @classmethod
    def _add_sub_module(cls, submodule):
        """
        Add a submodule.
        :param Text submodule:  This should be a single submodule. Do NOT include periods
        :rtype: LazyLoadModule
        """
        m = cls._lazy_submodules.get(submodule)
        if not m:
            m = cls._lazy_submodules[submodule] = lazy_load_module("{}.{}".format(cls._module, submodule))
        return m

    @classmethod
    def add_sub_module(cls, submodule):
        """
        Add a submodule.
        :param Text submodule: If periods are included, it will be added recursively
        :rtype: LazyLoadModule
        """
        parts = submodule.split(".", 1)
        m = cls._add_sub_module(parts[0])
        if len(parts) > 1:
            m = type(m).add_sub_module(parts[1])
        return m

    @classmethod
    def tag_with_plugin(cls, p: LazyLoadPlugin):
        """
        :param LazyLoadPlugin p:
        """
        cls._plugins.append(p)



import logging as _logging
import os as _os

import click

from flytekit.clis.sdk_in_container.constants import (
    CTX_DOMAIN,
    CTX_PACKAGES,
    CTX_PROJECT,
    CTX_TEST,
    CTX_VERSION,
    domain_option,
    project_option,
    version_option,
)
from flytekit.common import utils as _utils
from flytekit.common.core import identifier as _identifier
from flytekit.common.tasks import task as _task
from flytekit.configuration.internal import DOMAIN as _DOMAIN
from flytekit.configuration.internal import IMAGE as _IMAGE
from flytekit.configuration.internal import PROJECT as _PROJECT
from flytekit.configuration.internal import VERSION as _VERSION
from flytekit.configuration.internal import look_up_version_from_image_tag as _look_up_version_from_image_tag
from flytekit.tools.module_loader import iterate_registerable_entities_in_order


def register_all(project, domain, pkgs, test, version):
    if test:
        click.echo("Test switch enabled, not doing anything...")
    click.echo(
        "Running task, workflow, and launch plan registration for {}, {}, {} with version {}".format(
            project, domain, pkgs, version
        )
    )

    # m = module (i.e. python file)
    # k = value of dir(m), type str
    # o = object (e.g. SdkWorkflow)
    for m, k, o in iterate_registerable_entities_in_order(pkgs):
        name = _utils.fqdn(m.__name__, k, entity_type=o.resource_type)
        _logging.debug("Found module {}\n   K: {} Instantiated in {}".format(m, k, o._instantiated_in))
        o._id = _identifier.Identifier(o.resource_type, project, domain, name, version)

        if test:
            click.echo("Would register {:20} {}".format("{}:".format(o.entity_type_text), o.id.name))
        else:
            click.echo("Registering {:20} {}".format("{}:".format(o.entity_type_text), o.id.name))
            o.register(project, domain, o.id.name, version)


def register_tasks_only(project, domain, pkgs, test, version):
    if test:
        click.echo("Test switch enabled, not doing anything...")

    click.echo("Running task only registration for {}, {}, {} with version {}".format(project, domain, pkgs, version))

    # Discover all tasks by loading the module
    for m, k, t in iterate_registerable_entities_in_order(pkgs, include_entities={_task.SdkTask}):
        name = _utils.fqdn(m.__name__, k, entity_type=t.resource_type)

        if test:
            click.echo("Would register task {:20} {}".format("{}:".format(t.entity_type_text), name))
        else:
            click.echo("Registering task {:20} {}".format("{}:".format(t.entity_type_text), name))
            t.register(project, domain, name, version)


@click.group("register")
@project_option
@domain_option
@version_option
# --pkgs on the register group is DEPRECATED, use same arg on pyflyte.main instead
@click.option(
    "--pkgs",
    multiple=True,
    help="DEPRECATED. This arg can only be used before the 'register' keyword",
)
@click.option("--test", is_flag=True, help="Dry run, do not actually register with Admin")
@click.pass_context
def register(ctx, project, domain, version, pkgs=None, test=None):
    """
    Run registration steps for the workflows in this container.

    Run with the --test switch for a dry run to see what will be registered.  A default launch plan will also be
    created, if a role can be found in the environment variables.
    """
    if pkgs:
        raise click.UsageError("--pkgs must now be specified before the 'register' keyword on the command line")

    version = version or _look_up_version_from_image_tag(_IMAGE.get())
    if not version:
        raise click.UsageError("Could not find image from config, please specify a value for ``--version``")

    ctx.obj[CTX_PROJECT] = project
    ctx.obj[CTX_DOMAIN] = domain
    ctx.obj[CTX_VERSION] = version
    ctx.obj[CTX_TEST] = test
    _os.environ[_PROJECT.env_var] = project
    _os.environ[_DOMAIN.env_var] = domain
    _os.environ[_VERSION.env_var] = ctx.obj[CTX_VERSION]


@click.command("tasks")
@click.option(
    "-v",
    "--version",
    type=str,
    help="Version to register tasks with. This is normally parsed from the" "image, but you can override here.",
)
@click.pass_context
def tasks(ctx, version=None):
    """
    Only register tasks.
    """
    project = ctx.obj[CTX_PROJECT]
    domain = ctx.obj[CTX_DOMAIN]
    test = ctx.obj[CTX_TEST]
    pkgs = ctx.obj[CTX_PACKAGES]

    version = version or ctx.obj[CTX_VERSION] or _look_up_version_from_image_tag(_IMAGE.get())
    register_tasks_only(project, domain, pkgs, test, version)


@click.command("workflows")
@click.option(
    "-v",
    "--version",
    type=str,
    help="Version to register tasks with. This is normally parsed from the" "image, but you can override here.",
)
@click.pass_context
def workflows(ctx, version=None):
    """
    Register both tasks and workflows.  Also create and register a default launch plan for all workflows.
    """
    project = ctx.obj[CTX_PROJECT]
    domain = ctx.obj[CTX_DOMAIN]
    test = ctx.obj[CTX_TEST]
    pkgs = ctx.obj[CTX_PACKAGES]

    version = version or ctx.obj[CTX_VERSION] or _look_up_version_from_image_tag(_IMAGE.get())
    register_all(project, domain, pkgs, test, version)


register.add_command(tasks)
register.add_command(workflows)



import six as _six

from flytekit.common import sdk_bases as _sdk_bases
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.models.core import identifier as _core_identifier


class Identifier(_core_identifier.Identifier, metaclass=_sdk_bases.ExtendedSdkType):

    _STRING_TO_TYPE_MAP = {
        "lp": _core_identifier.ResourceType.LAUNCH_PLAN,
        "wf": _core_identifier.ResourceType.WORKFLOW,
        "tsk": _core_identifier.ResourceType.TASK,
    }
    _TYPE_TO_STRING_MAP = {v: k for k, v in _six.iteritems(_STRING_TO_TYPE_MAP)}

    @classmethod
    def promote_from_model(cls, base_model):
        """
        :param flytekit.models.core.identifier.Identifier base_model:
        :rtype: Identifier
        """
        return cls(
            base_model.resource_type,
            base_model.project,
            base_model.domain,
            base_model.name,
            base_model.version,
        )

    @classmethod
    def from_python_std(cls, string):
        """
        Parses a string in the correct format into an identifier
        :param Text string:
        :rtype: Identifier
        """
        segments = string.split(":")
        if len(segments) != 5:
            raise _user_exceptions.FlyteValueException(
                "The provided string was not in a parseable format. The string for an identifier must be in the format"
                " entity_type:project:domain:name:version.  Received: {}".format(string)
            )

        resource_type, project, domain, name, version = segments

        if resource_type not in cls._STRING_TO_TYPE_MAP:
            raise _user_exceptions.FlyteValueException(
                "The provided string could not be parsed. The first element of an identifier must be one of: {}. "
                "Received: {}".format(list(cls._STRING_TO_TYPE_MAP.keys()), resource_type)
            )
        resource_type = cls._STRING_TO_TYPE_MAP[resource_type]

        return cls(resource_type, project, domain, name, version)

    def __str__(self):
        return "{}:{}:{}:{}:{}".format(
            type(self)._TYPE_TO_STRING_MAP.get(self.resource_type, "<unknown>"),
            self.project,
            self.domain,
            self.name,
            self.version,
        )


class WorkflowExecutionIdentifier(_core_identifier.WorkflowExecutionIdentifier, metaclass=_sdk_bases.ExtendedSdkType):
    @classmethod
    def promote_from_model(cls, base_model):
        """
        :param flytekit.models.core.identifier.WorkflowExecutionIdentifier base_model:
        :rtype: WorkflowExecutionIdentifier
        """
        return cls(
            base_model.project,
            base_model.domain,
            base_model.name,
        )

    @classmethod
    def from_python_std(cls, string):
        """
        Parses a string in the correct format into an identifier
        :param Text string:
        :rtype: WorkflowExecutionIdentifier
        """
        segments = string.split(":")
        if len(segments) != 4:
            raise _user_exceptions.FlyteValueException(
                string,
                "The provided string was not in a parseable format. The string for an identifier must be in the format"
                " ex:project:domain:name.",
            )

        resource_type, project, domain, name = segments

        if resource_type != "ex":
            raise _user_exceptions.FlyteValueException(
                resource_type,
                "The provided string could not be parsed. The first element of an execution identifier must be 'ex'.",
            )

        return cls(
            project,
            domain,
            name,
        )

    def __str__(self):
        return "ex:{}:{}:{}".format(self.project, self.domain, self.name)


class TaskExecutionIdentifier(_core_identifier.TaskExecutionIdentifier, metaclass=_sdk_bases.ExtendedSdkType):
    @classmethod
    def promote_from_model(cls, base_model):
        """
        :param flytekit.models.core.identifier.TaskExecutionIdentifier base_model:
        :rtype: TaskExecutionIdentifier
        """
        return cls(
            task_id=base_model.task_id,
            node_execution_id=base_model.node_execution_id,
            retry_attempt=base_model.retry_attempt,
        )

    @classmethod
    def from_python_std(cls, string):
        """
        Parses a string in the correct format into an identifier
        :param Text string:
        :rtype: TaskExecutionIdentifier
        """
        segments = string.split(":")
        if len(segments) != 10:
            raise _user_exceptions.FlyteValueException(
                string,
                "The provided string was not in a parseable format. The string for an identifier must be in the format"
                " te:exec_project:exec_domain:exec_name:node_id:task_project:task_domain:task_name:task_version:retry.",
            )

        resource_type, ep, ed, en, node_id, tp, td, tn, tv, retry = segments

        if resource_type != "te":
            raise _user_exceptions.FlyteValueException(
                resource_type,
                "The provided string could not be parsed. The first element of an execution identifier must be 'ex'.",
            )

        return cls(
            task_id=Identifier(_core_identifier.ResourceType.TASK, tp, td, tn, tv),
            node_execution_id=_core_identifier.NodeExecutionIdentifier(
                node_id=node_id,
                execution_id=_core_identifier.WorkflowExecutionIdentifier(ep, ed, en),
            ),
            retry_attempt=int(retry),
        )

    def __str__(self):
        return "te:{ep}:{ed}:{en}:{node_id}:{tp}:{td}:{tn}:{tv}:{retry}".format(
            ep=self.node_execution_id.execution_id.project,
            ed=self.node_execution_id.execution_id.domain,
            en=self.node_execution_id.execution_id.name,
            node_id=self.node_execution_id.node_id,
            tp=self.task_id.project,
            td=self.task_id.domain,
            tn=self.task_id.name,
            tv=self.task_id.version,
            retry=self.retry_attempt,
        )



from wrapt import decorator as _decorator

from flytekit.common import utils as _utils
from flytekit.interfaces.data import data_proxy as _data_proxy


class LocalTestFileSystem(object):
    """
    Context manager for creating a temporary test file system locally for the purpose of unit testing and grabbing
    remote objects.  This scratch space will be automatically cleaned up as long as sys.exit() is not called from
    within the context.  This context need only be used in user scripts and tests--all task executions are guaranteed
    to have the necessary managed disk context available.

    .. note::

        This is especially useful when dealing with remote blob-like objects (Blob, CSV, MultiPartBlob,
        MultiPartCSV, Schema) as they require backing on disk.  Using this context manager creates that disk context
        to support the downloads.

    .. note::

        Blob-like objects can be downloaded to user-specified locations.  See documentation for
        flytekit.sdk.types.Types for more information.

    .. note::

        When this context is entered, it overrides any contexts already entered.  All blobs will be written to the most
        recent entered context. Upon exiting the context, all data associated will be deleted.  It is recommended to
        only use one LocalTestFileSystem() per test to avoid confusion.

    .. code-block:: python

        with LocalTestFileSystem():
            wf_handle = SdkWorkflowExecution.fetch('project', 'domain', 'name')
            with wf_handle.node_executions['my_node'].outputs.blob as reader:
                assert reader.read() == "hello!"
    """

    def __init__(self):
        self._exit_stack = _utils.ExitStack()

    def __enter__(self):
        """
        :rtype: flytekit.common.utils.AutoDeletingTempDir
        """
        self._exit_stack.__enter__()
        temp_dir = self._exit_stack.enter_context(_utils.AutoDeletingTempDir("local_test_filesystem"))
        self._exit_stack.enter_context(_data_proxy.LocalDataContext(temp_dir.name))
        self._exit_stack.enter_context(_data_proxy.LocalWorkingDirectoryContext(temp_dir))
        return temp_dir

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._exit_stack.__exit__(exc_type, exc_val, exc_tb)


@_decorator
def flyte_test(fn, _, args, kwargs):
    """
    This is a decorator which can be used to annotate test functions.  By using this decorator, the necessary local
    scratch context will be prepared and then cleaned up upon completion.

    .. code-block:: python

        @inputs(input_blob=Types.Blob)
        @outputs(response_blob=Types.Blob)
        @python_task
        def test_task(wf_params, input_blob, response_blob):
            response = Types.Blob()
            with response as writer:
                with input_blob as reader:
                    txt = reader.read()
                    if txt == "Hi":
                        writer.write("Hello, world!")
                    elif txt == "Goodnight":
                        writer.write("Goodnight, moon.")
                    else:
                        writer.write("Does not compute".)
            response_blob.set(response)

        @flyte_test
        def some_test():
            blob = Types.Blob()
            with blob as writer:
                writer.write("Hi")

            result = test_task.unit_test(input_blob=blob)

            with result['response_blob'] as reader:
                assert reader.read() == 'Hello, world!"
    """
    with LocalTestFileSystem():
        return fn(*args, **kwargs)



from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.types import base_sdk_types as _base_sdk_types
from flytekit.common.types.impl import schema as _schema_impl
from flytekit.models import literals as _literals
from flytekit.models import types as _idl_types


class SchemaInstantiator(_base_sdk_types.InstantiableType):
    def create_at_known_location(cls, location):
        """
        :param Text location:
        :rtype: flytekit.common.types.impl.schema.Schema
        """
        return _schema_impl.Schema.create_at_known_location(location, mode="wb", schema_type=cls.schema_type)

    def fetch(cls, remote_path, local_path=None):
        """
        :param Text remote_path:
        :param Text local_path: [Optional] If specified, the Schema is copied to this location.  If specified,
            this location is NOT managed and the schema will not be cleaned up upon exit.
        :rtype: flytekit.common.types.impl.schema.Schema
        """
        return _schema_impl.Schema.fetch(remote_path, mode="rb", local_path=local_path, schema_type=cls.schema_type)

    def create(cls):
        """
        :rtype: flytekit.common.types.impl.schema.Schema
        """
        return _schema_impl.Schema.create_at_any_location(mode="wb", schema_type=cls.schema_type)

    def create_from_hive_query(
        cls,
        select_query,
        stage_query=None,
        schema_to_table_name_map=None,
        known_location=None,
    ):
        """
        Returns a query that can be submitted to Hive and produce the desired output.  It also returns a properly-typed
        schema object.

        :param Text select_query: Query for selecting data from Hive
        :param Text stage_query: Query for building temporary tables on Hive.
            Runs before the select query. Temporary tables are supported but CTEs are not supported.
        :param Dict[Text, Text] schema_to_table_name_map: A map of column names in the schema to the column names
            returned from the select query
        :param Text known_location: create the schema object at a known s3 location.
        :rtype: flytekit.common.types.impl.schema.Schema, Text
        """
        return _schema_impl.Schema.create_from_hive_query(
            select_query=select_query,
            stage_query=stage_query,
            schema_to_table_name_map=schema_to_table_name_map,
            known_location=known_location,
            schema_type=cls.schema_type,
        )

    def __call__(cls, *args, **kwargs):
        """
        TODO: Is there a better way to deal with this?

        :rtype: flytekit.common.types.impl.schema.Schema
        """
        if not args and not kwargs:
            return _schema_impl.Schema.create_at_any_location(mode="wb", schema_type=cls.schema_type)
        else:
            return super(SchemaInstantiator, cls).__call__(*args, **kwargs)

    @property
    def schema_type(cls):
        """
        :rtype: _schema_impl.SchemaType
        """
        return cls._schema_type

    @property
    def columns(cls):
        """
        :rtype: dict[Text, flytekit.common.types.base_sdk_types.FlyteSdkType]
        """
        return cls.schema_type.sdk_columns


class Schema(_base_sdk_types.FlyteSdkValue, metaclass=SchemaInstantiator):
    @classmethod
    def from_string(cls, string_value):
        """
        :param Text string_value:
        :rtype: Schema
        """
        if not string_value:
            _user_exceptions.FlyteValueException(string_value, "Cannot create a Schema from an empty path")
        return cls(_schema_impl.Schema.from_string(string_value, schema_type=cls.schema_type))

    @classmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        return cls == other

    @classmethod
    def from_python_std(cls, t_value):
        """
        :param T t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        if t_value is None:
            return _base_sdk_types.Void()
        elif isinstance(t_value, _schema_impl.Schema):
            schema = t_value.cast_to(cls.schema_type)
        else:
            schema = _schema_impl.Schema.from_python_std(t_value, schema_type=cls.schema_type)
        return cls(schema)

    @classmethod
    def to_flyte_literal_type(cls):
        """
        :rtype: flytekit.models.types.LiteralType
        """
        return _idl_types.LiteralType(schema=cls.schema_type)

    @classmethod
    def promote_from_model(cls, literal_model):
        """
        Creates an object of this type from the model primitive defining it.
        :param flytekit.models.literals.Literal literal_model:
        :rtype: Schema
        """
        return cls(_schema_impl.Schema.promote_from_model(literal_model.scalar.schema))

    @classmethod
    def short_class_string(cls):
        """
        :rtype: Text
        """
        return repr(cls.schema_type)

    def __init__(self, value):
        """
        :param flytekit.common.types.impl.schema.Schema value: Schema value to wrap
        """
        super(Schema, self).__init__(scalar=_literals.Scalar(schema=value))

    def to_python_std(self):
        """
        :rtype: flytekit.common.types.impl.schema.Schema
        """
        return self.scalar.schema

    def short_string(self):
        """
        :rtype: Text
        """
        return "{}".format(
            self.scalar.schema,
        )


def schema_instantiator(columns=None):
    """
    :param list[(Text, flytekit.common.types.base_sdk_types.FlyteSdkType)] columns: [Optional] Description of the
        columns in the underlying schema.  Should be tuples with the first element being the name.
    :rtype: SchemaInstantiator
    """
    if columns is not None and len(columns) == 0:
        raise _user_exceptions.FlyteValueException(
            columns,
            "When specifying a Schema type with a known set of columns, a non-empty list must be provided as " "inputs",
        )

    class _Schema(Schema, metaclass=SchemaInstantiator):
        _schema_type = _schema_impl.SchemaType(columns=columns)

    return _Schema


def schema_instantiator_from_proto(schema_type):
    """
    :param flytekit.models.types.SchemaType schema_type:
    :rtype: SchemaInstantiator
    """

    class _Schema(Schema, metaclass=SchemaInstantiator):
        _schema_type = _schema_impl.SchemaType.promote_from_model(schema_type)

    return _Schema



import requests as _requests

from flytekit.common.exceptions import user as _user_exceptions
from flytekit.interfaces.data import common as _common_data


class HttpFileProxy(_common_data.DataProxy):

    _HTTP_OK = 200
    _HTTP_FORBIDDEN = 403
    _HTTP_NOT_FOUND = 404

    def exists(self, path):
        """
        :param Text path: the path of the file
        :rtype bool: whether the file exists or not
        """
        rsp = _requests.head(path)
        allowed_codes = {
            type(self)._HTTP_OK,
            type(self)._HTTP_NOT_FOUND,
            type(self)._HTTP_FORBIDDEN,
        }
        if rsp.status_code not in allowed_codes:
            raise _user_exceptions.FlyteValueException(
                rsp.status_code,
                "Data at {} could not be checked for existence. Expected one of: {}".format(path, allowed_codes),
            )
        return rsp.status_code == type(self)._HTTP_OK

    def download_directory(self, from_path, to_path):
        """
        :param Text from_path:
        :param Text to_path:
        """
        raise _user_exceptions.FlyteAssertion("Reading data recursively from HTTP endpoint is not currently supported.")

    def download(self, from_path, to_path):
        """
        :param Text from_path:
        :param Text to_path:
        """

        rsp = _requests.get(from_path)
        if rsp.status_code != type(self)._HTTP_OK:
            raise _user_exceptions.FlyteValueException(
                rsp.status_code,
                "Request for data @ {} failed. Expected status code {}".format(from_path, type(self)._HTTP_OK),
            )
        with open(to_path, "wb") as writer:
            writer.write(rsp.content)

    def upload(self, from_path, to_path):
        """
        :param Text from_path:
        :param Text to_path:
        """
        raise _user_exceptions.FlyteAssertion("Writing data to HTTP endpoint is not currently supported.")

    def upload_directory(self, from_path, to_path):
        """
        :param Text from_path:
        :param Text to_path:
        """
        raise _user_exceptions.FlyteAssertion("Writing data to HTTP endpoint is not currently supported.")

    def get_random_path(self):
        """
        :rtype: Text
        """
        raise _user_exceptions.FlyteAssertion("Writing data to HTTP endpoint is not currently supported.")

    def get_random_directory(self):
        """
        :rtype: Text
        """
        raise _user_exceptions.FlyteAssertion("Writing data to HTTP endpoint is not currently supported.")



import contextlib
import os
import shutil
import sqlite3
import tempfile
import typing
from dataclasses import dataclass

import pandas as pd

from flytekit import FlyteContext, kwtypes
from flytekit.core.base_sql_task import SQLTask
from flytekit.core.python_function_task import PythonInstanceTask
from flytekit.types.schema import FlyteSchema


def unarchive_file(local_path: str, to_dir: str):
    """
    Unarchive given archive and returns the unarchived file name. It is expected that only one file is unarchived.
    More than one file or 0 files will result in a ``RuntimeError``
    """
    archive_dir = os.path.join(to_dir, "_arch")
    shutil.unpack_archive(local_path, archive_dir)
    # file gets uncompressed into to_dir/_arch/*.*
    files = os.listdir(archive_dir)
    if not files or len(files) == 0 or len(files) > 1:
        raise RuntimeError(f"Uncompressed archive should contain only one file - found {files}!")
    return os.path.join(archive_dir, files[0])


@dataclass
class SQLite3Config(object):
    """
    Use this configuration to configure if sqlite3 files that should be loaded by the task. The file itself is
    considered as a database and hence is treated like a configuration
    The path to a static sqlite3 compatible database file can be
      - within the container
      - or from a publicly downloadable source

    Args:
        uri: default FlyteFile that will be downloaded on execute
        compressed: Boolean that indicates if the given file is a compressed archive. Supported file types are
                        [zip, tar, gztar, bztar, xztar]
    """

    uri: str
    compressed: bool = False


class SQLite3Task(PythonInstanceTask[SQLite3Config], SQLTask[SQLite3Config]):
    """
    Makes it possible to run client side SQLite3 queries that optionally return a FlyteSchema object

    TODO: How should we use pre-built containers for running portable tasks like this. Should this always be a
          referenced task type?
    """

    _SQLITE_TASK_TYPE = "sqlite"

    def __init__(
        self,
        name: str,
        query_template: str,
        inputs: typing.Optional[typing.Dict[str, typing.Type]] = None,
        task_config: typing.Optional[SQLite3Config] = None,
        output_schema_type: typing.Optional[typing.Type[FlyteSchema]] = None,
        **kwargs,
    ):
        if task_config is None or task_config.uri is None:
            raise ValueError("SQLite DB uri is required.")
        outputs = kwtypes(results=output_schema_type if output_schema_type else FlyteSchema)
        super().__init__(
            name=name,
            task_config=task_config,
            task_type=self._SQLITE_TASK_TYPE,
            query_template=query_template,
            inputs=inputs,
            outputs=outputs,
            **kwargs,
        )

    @property
    def output_columns(self) -> typing.Optional[typing.List[str]]:
        c = self.python_interface.outputs["results"].column_names()
        return c if c else None

    def execute(self, **kwargs) -> typing.Any:
        with tempfile.TemporaryDirectory() as temp_dir:
            ctx = FlyteContext.current_context()
            file_ext = os.path.basename(self.task_config.uri)
            local_path = os.path.join(temp_dir, file_ext)
            ctx.file_access.download(self.task_config.uri, local_path)
            if self.task_config.compressed:
                local_path = unarchive_file(local_path, temp_dir)

            print(f"Connecting to db {local_path}")
            with contextlib.closing(sqlite3.connect(local_path)) as con:
                df = pd.read_sql_query(self.get_query(**kwargs), con)
                return df



import logging as _logging

from flytekit.common import sdk_bases as _sdk_bases
from flytekit.common.exceptions import system as _system_exceptions
from flytekit.models.core import workflow as _workflow_model


class SdkTaskNode(_workflow_model.TaskNode, metaclass=_sdk_bases.ExtendedSdkType):
    def __init__(self, sdk_task):
        """
        :param flytekit.common.tasks.task.SdkTask sdk_task:
        """
        self._sdk_task = sdk_task
        super(SdkTaskNode, self).__init__(None)

    @property
    def reference_id(self):
        """
        A globally unique identifier for the task.
        :rtype: flytekit.models.core.identifier.Identifier
        """
        return self._sdk_task.id

    @property
    def sdk_task(self):
        """
        :rtype: flytekit.common.tasks.task.SdkTask
        """
        return self._sdk_task

    @classmethod
    def promote_from_model(cls, base_model, tasks):
        """
        Takes the idl wrapper for a TaskNode and returns the hydrated Flytekit object for it by fetching it from the
        engine.

        :param flytekit.models.core.workflow.TaskNode base_model:
        :param dict[flytekit.models.core.identifier.Identifier, flytekit.models.task.TaskTemplate] tasks:
        :rtype: SdkTaskNode
        """
        from flytekit.common.tasks import task as _task

        if base_model.reference_id in tasks:
            t = tasks[base_model.reference_id]
            _logging.info(f"Found existing task template for {t.id}, will not retrieve from Admin")
            sdk_task = _task.SdkTask.promote_from_model(t)
            sdk_task._has_registered = True
            return cls(sdk_task)

        # If not found, fetch it from Admin
        _logging.debug("Fetching task template for {} from Admin".format(base_model.reference_id))
        project = base_model.reference_id.project
        domain = base_model.reference_id.domain
        name = base_model.reference_id.name
        version = base_model.reference_id.version
        sdk_task = _task.SdkTask.fetch(project, domain, name, version)
        return cls(sdk_task)


class SdkWorkflowNode(_workflow_model.WorkflowNode, metaclass=_sdk_bases.ExtendedSdkType):
    def __init__(self, sdk_workflow=None, sdk_launch_plan=None):
        """
        :param flytekit.common.workflow.SdkWorkflow sdk_workflow:
        :param flytekit.common.launch_plan.SdkLaunchPlan sdk_launch_plan:
        """
        if sdk_workflow and sdk_launch_plan:
            raise _system_exceptions.FlyteSystemException(
                "SdkWorkflowNode cannot be called with both a workflow and "
                "a launchplan specified, please pick one. WF: {} LP: {}",
                sdk_workflow,
                sdk_launch_plan,
            )

        self._sdk_workflow = sdk_workflow
        self._sdk_launch_plan = sdk_launch_plan
        sdk_wf_id = sdk_workflow.id if sdk_workflow else None
        sdk_lp_id = sdk_launch_plan.id if sdk_launch_plan else None
        super(SdkWorkflowNode, self).__init__(launchplan_ref=sdk_lp_id, sub_workflow_ref=sdk_wf_id)

    def __repr__(self):
        """
        :rtype: Text
        """
        if self.sdk_workflow is not None:
            return "SdkWorkflowNode with workflow: {}".format(self.sdk_workflow)
        return "SdkWorkflowNode with launch plan: {}".format(self.sdk_launch_plan)

    @property
    def launchplan_ref(self):
        """
        [Optional] A globally unique identifier for the launch plan.  Should map to Admin.
        :rtype: flytekit.models.core.identifier.Identifier
        """
        return self._sdk_launch_plan.id if self._sdk_launch_plan else None

    @property
    def sub_workflow_ref(self):
        """
        [Optional] Reference to a subworkflow, that should be defined with the compiler context.
        :rtype: flytekit.models.core.identifier.Identifier
        """
        return self._sdk_workflow.id if self._sdk_workflow else None

    @property
    def sdk_launch_plan(self):
        """
        :rtype: flytekit.common.launch_plan.SdkLaunchPlan
        """
        return self._sdk_launch_plan

    @property
    def sdk_workflow(self):
        """
        :rtype: flytekit.common.workflow.SdkWorkflow
        """
        return self._sdk_workflow

    @classmethod
    def promote_from_model(cls, base_model, sub_workflows, tasks):
        """
        :param flytekit.models.core.workflow.WorkflowNode base_model:
        :param dict[flytekit.models.core.identifier.Identifier, flytekit.models.core.workflow.WorkflowTemplate]
            sub_workflows:
        :param dict[flytekit.models.core.identifier.Identifier, flytekit.models.task.TaskTemplate] tasks:
        :rtype: SdkWorkflowNode
        """
        # put the import statement here to prevent circular dependency error
        from flytekit.common import launch_plan as _launch_plan
        from flytekit.common import workflow as _workflow

        project = base_model.reference.project
        domain = base_model.reference.domain
        name = base_model.reference.name
        version = base_model.reference.version
        if base_model.launchplan_ref is not None:
            sdk_launch_plan = _launch_plan.SdkLaunchPlan.fetch(project, domain, name, version)
            return cls(sdk_launch_plan=sdk_launch_plan)
        elif base_model.sub_workflow_ref is not None:
            # The workflow templates for sub-workflows should have been included in the original response
            if base_model.reference in sub_workflows:
                sw = sub_workflows[base_model.reference]
                promoted = _workflow.SdkWorkflow.promote_from_model(sw, sub_workflows=sub_workflows, tasks=tasks)
                return cls(sdk_workflow=promoted)

            # If not found for some reason, fetch it from Admin again.
            # The reason there is a warning here but not for tasks is because sub-workflows should always be passed
            # along. Ideally subworkflows are never even registered with Admin, so fetching from Admin ideally doesn't
            # return anything.
            _logging.warning(
                "Your subworkflow with id {} is not included in the promote call.".format(base_model.reference)
            )
            sdk_workflow = _workflow.SdkWorkflow.fetch(project, domain, name, version)
            return cls(sdk_workflow=sdk_workflow)
        else:
            raise _system_exceptions.FlyteSystemException(
                "Bad workflow node model, neither subworkflow nor " "launchplan specified."
            )



from enum import Enum
from typing import Any, Dict, List, Optional, Type

from flytekit.common.tasks.raw_container import _get_container_definition
from flytekit.core.base_task import PythonTask, TaskMetadata
from flytekit.core.context_manager import SerializationSettings
from flytekit.core.interface import Interface
from flytekit.core.resources import Resources, ResourceSpec
from flytekit.models import task as _task_model


class ContainerTask(PythonTask):
    class MetadataFormat(Enum):
        JSON = _task_model.DataLoadingConfig.LITERALMAP_FORMAT_JSON
        YAML = _task_model.DataLoadingConfig.LITERALMAP_FORMAT_YAML
        PROTO = _task_model.DataLoadingConfig.LITERALMAP_FORMAT_PROTO

    class IOStrategy(Enum):
        DOWNLOAD_EAGER = _task_model.IOStrategy.DOWNLOAD_MODE_EAGER
        DOWNLOAD_STREAM = _task_model.IOStrategy.DOWNLOAD_MODE_STREAM
        DO_NOT_DOWNLOAD = _task_model.IOStrategy.DOWNLOAD_MODE_NO_DOWNLOAD
        UPLOAD_EAGER = _task_model.IOStrategy.UPLOAD_MODE_EAGER
        UPLOAD_ON_EXIT = _task_model.IOStrategy.UPLOAD_MODE_ON_EXIT
        DO_NOT_UPLOAD = _task_model.IOStrategy.UPLOAD_MODE_NO_UPLOAD

    def __init__(
        self,
        name: str,
        image: str,
        command: List[str],
        inputs: Optional[Dict[str, Type]] = None,
        metadata: Optional[TaskMetadata] = None,
        arguments: List[str] = None,
        outputs: Dict[str, Type] = None,
        requests: Optional[Resources] = None,
        limits: Optional[Resources] = None,
        input_data_dir: str = None,
        output_data_dir: str = None,
        metadata_format: MetadataFormat = MetadataFormat.JSON,
        io_strategy: IOStrategy = None,
        **kwargs,
    ):
        super().__init__(
            task_type="raw-container",
            name=name,
            interface=Interface(inputs, outputs),
            metadata=metadata,
            task_config=None,
            **kwargs,
        )
        self._image = image
        self._cmd = command
        self._args = arguments
        self._input_data_dir = input_data_dir
        self._output_data_dir = output_data_dir
        self._md_format = metadata_format
        self._io_strategy = io_strategy
        self._resources = ResourceSpec(
            requests=requests if requests else Resources(), limits=limits if limits else Resources()
        )

    @property
    def resources(self) -> ResourceSpec:
        return self._resources

    def execute(self, **kwargs) -> Any:
        print(kwargs)
        env = ""
        for k, v in self.environment.items():
            env += f" -e {k}={v}"
        print(
            f"\ndocker run --rm -v /tmp/inputs:{self._input_data_dir} -v /tmp/outputs:{self._output_data_dir} {env}"
            f"{self._image} {self._cmd} {self._args}"
        )
        return None

    def get_container(self, settings: SerializationSettings) -> _task_model.Container:
        env = {**settings.env, **self.environment} if self.environment else settings.env
        return _get_container_definition(
            image=self._image,
            command=self._cmd,
            args=self._args,
            data_loading_config=_task_model.DataLoadingConfig(
                input_path=self._input_data_dir,
                output_path=self._output_data_dir,
                format=self._md_format.value,
                enabled=True,
                io_strategy=self._io_strategy.value if self._io_strategy else None,
            ),
            environment=env,
            cpu_request=self.resources.requests.cpu,
            cpu_limit=self.resources.limits.cpu,
            memory_request=self.resources.requests.mem,
            memory_limit=self.resources.limits.mem,
        )



import logging as _logging
import os as _os
import traceback as _traceback
from datetime import datetime as _datetime

import six as _six
from deprecated import deprecated as _deprecated
from flyteidl.core import literals_pb2 as _literals_pb2

import flytekit
from flytekit.clients.friendly import SynchronousFlyteClient as _SynchronousFlyteClient
from flytekit.clients.helpers import iterate_node_executions as _iterate_node_executions
from flytekit.clients.helpers import iterate_task_executions as _iterate_task_executions
from flytekit.common import constants as _constants
from flytekit.common import utils as _common_utils
from flytekit.common.exceptions import scopes as _exception_scopes
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.configuration import auth as _auth_config
from flytekit.configuration import internal as _internal_config
from flytekit.configuration import platform as _platform_config
from flytekit.configuration import sdk as _sdk_config
from flytekit.engines import common as _common_engine
from flytekit.interfaces.data import data_proxy as _data_proxy
from flytekit.interfaces.stats.taggable import get_stats as _get_stats
from flytekit.models import common as _common_models
from flytekit.models import execution as _execution_models
from flytekit.models import literals as _literals
from flytekit.models import task as _task_models
from flytekit.models.admin import common as _common
from flytekit.models.admin import workflow as _workflow_model
from flytekit.models.core import errors as _error_models
from flytekit.models.core import identifier as _identifier


class _FlyteClientManager(object):
    _CLIENT = None

    def __init__(self, *args, **kwargs):
        # TODO: React to changing configs.  For now this is frozen for the lifetime of the process, which covers most
        # TODO: use cases.
        if type(self)._CLIENT is None:
            c = _SynchronousFlyteClient(*args, **kwargs)
            type(self)._CLIENT = c

    @property
    def client(self):
        """
        :rtype: flytekit.clients.friendly.SynchronousFlyteClient
        """
        return type(self)._CLIENT


# This is a simple helper function that ties the client together with the configuration construct.
# This will be refactored away when we move to a heavier context object.
def get_client() -> _SynchronousFlyteClient:
    return _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client


class FlyteEngineFactory(_common_engine.BaseExecutionEngineFactory):
    def get_workflow(self, sdk_workflow):
        """
        :param flytekit.common.workflow.SdkWorkflow sdk_workflow:
        :rtype: FlyteWorkflow
        """
        return FlyteWorkflow(sdk_workflow)

    def get_task(self, sdk_task):
        """
        :param flytekit.common.tasks.task.SdkTask sdk_task:
        :rtype: FlyteTask
        """
        return FlyteTask(sdk_task)

    def get_launch_plan(self, sdk_launch_plan):
        """
        :param flytekit.common.launch_plan.SdkLaunchPlan sdk_launch_plan:
        :rtype: FlyteLaunchPlan
        """
        return FlyteLaunchPlan(sdk_launch_plan)

    def get_task_execution(self, task_exec):
        """
        :param flytekit.common.tasks.executions.SdkTaskExecution task_exec:
        :rtype: FlyteTaskExecution
        """
        return FlyteTaskExecution(task_exec)

    def get_node_execution(self, node_exec):
        """
        :param flytekit.common.nodes.SdkNodeExecution node_exec:
        :rtype: FlyteNodeExecution
        """
        return FlyteNodeExecution(node_exec)

    def get_workflow_execution(self, wf_exec):
        """
        :param flytekit.common.workflow_execution.SdkWorkflowExecution wf_exec:
        :rtype: FlyteWorkflowExecution
        """
        return FlyteWorkflowExecution(wf_exec)

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def fetch_workflow_execution(self, wf_exec_id):
        """
        :param flytekit.models.core.identifier.WorkflowExecutionIdentifier wf_exec_id:
        :rtype: flytekit.models.execution.Execution
        """
        return _FlyteClientManager(
            _platform_config.URL.get(), insecure=_platform_config.INSECURE.get()
        ).client.get_execution(wf_exec_id)

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def fetch_task(self, task_id):
        """
        Queries Admin for an existing Admin task
        :param flytekit.models.core.identifier.Identifier task_id:
        :rtype: flytekit.models.task.Task
        """
        return _FlyteClientManager(
            _platform_config.URL.get(), insecure=_platform_config.INSECURE.get()
        ).client.get_task(task_id)

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def fetch_latest_task(self, named_task):
        """
        Fetches the latest task
        :param flytekit.models.common.NamedEntityIdentifier named_task: NamedEntityIdentifier to fetch
        :rtype: flytekit.models.task.Task
        """
        task_list, _ = _FlyteClientManager(
            _platform_config.URL.get(), insecure=_platform_config.INSECURE.get()
        ).client.list_tasks_paginated(
            named_task,
            limit=1,
            sort_by=_common.Sort("created_at", _common.Sort.Direction.DESCENDING),
        )
        return task_list[0] if task_list else None

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def fetch_launch_plan(self, launch_plan_id):
        """
        :param flytekit.models.core.identifier.Identifier launch_plan_id: This identifier should have a resource
            type of kind LaunchPlan.
        :rtype: flytekit.models.launch_plan.LaunchPlan
        """
        if launch_plan_id.version:
            return _FlyteClientManager(
                _platform_config.URL.get(), insecure=_platform_config.INSECURE.get()
            ).client.get_launch_plan(launch_plan_id)
        else:
            named_entity_id = _common_models.NamedEntityIdentifier(
                launch_plan_id.project, launch_plan_id.domain, launch_plan_id.name
            )
            return _FlyteClientManager(
                _platform_config.URL.get(), insecure=_platform_config.INSECURE.get()
            ).client.get_active_launch_plan(named_entity_id)

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def fetch_workflow(self, workflow_id):
        """
        :param flytekit.models.core.identifier.Identifier workflow_id: This identifier should have a resource
            type of kind LaunchPlan.
        :rtype: flytekit.models.admin.workflow.Workflow
        """
        return _FlyteClientManager(
            _platform_config.URL.get(), insecure=_platform_config.INSECURE.get()
        ).client.get_workflow(workflow_id)


class FlyteLaunchPlan(_common_engine.BaseLaunchPlanLauncher):
    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def register(self, identifier):
        client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
        try:
            client.create_launch_plan(identifier, self.sdk_launch_plan)
        except _user_exceptions.FlyteEntityAlreadyExistsException:
            pass

    @_deprecated(reason="Use launch instead", version="0.9.0")
    def execute(
        self,
        project,
        domain,
        name,
        inputs,
        notification_overrides=None,
        label_overrides=None,
        annotation_overrides=None,
    ):
        """
        Deprecated. Use launch instead.
        """
        return self.launch(
            project,
            domain,
            name,
            inputs,
            notification_overrides,
            label_overrides,
            annotation_overrides,
        )

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def launch(
        self,
        project,
        domain,
        name,
        inputs,
        notification_overrides=None,
        label_overrides=None,
        annotation_overrides=None,
    ):
        """
        Creates a workflow execution using parameters specified in the launch plan.
        :param Text project:
        :param Text domain:
        :param Text name:
        :param flytekit.models.literals.LiteralMap inputs:
        :param list[flytekit.models.common.Notification] notification_overrides: If specified, override the
            notifications.
        :param flytekit.models.common.Labels label_overrides:
        :param flytekit.models.common.Annotations annotation_overrides:
        :rtype: flytekit.models.execution.Execution
        """
        disable_all = notification_overrides == []
        if disable_all:
            notification_overrides = None
        else:
            notification_overrides = _execution_models.NotificationList(notification_overrides or [])
            disable_all = None

        try:
            client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
            exec_id = client.create_execution(
                project,
                domain,
                name,
                _execution_models.ExecutionSpec(
                    self.sdk_launch_plan.id,
                    _execution_models.ExecutionMetadata(
                        _execution_models.ExecutionMetadata.ExecutionMode.MANUAL,
                        "sdk",  # TODO: get principle
                        0,  # TODO: Detect nesting
                    ),
                    notifications=notification_overrides,
                    disable_all=disable_all,
                    labels=label_overrides,
                    annotations=annotation_overrides,
                ),
                inputs,
            )
        except _user_exceptions.FlyteEntityAlreadyExistsException:
            exec_id = _identifier.WorkflowExecutionIdentifier(project, domain, name)
        return client.get_execution(exec_id)

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def update(self, identifier, state):
        """
        :param flytekit.models.core.identifier.Identifier identifier: Identifier for launch plan to update
        :param int state: Enum value from flytekit.models.launch_plan.LaunchPlanState
        """
        return _FlyteClientManager(
            _platform_config.URL.get(), insecure=_platform_config.INSECURE.get()
        ).client.update_launch_plan(identifier, state)


class FlyteWorkflow(_common_engine.BaseWorkflowExecutor):
    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def register(self, identifier):
        client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
        try:
            sub_workflows = self.sdk_workflow.get_sub_workflows()
            return client.create_workflow(
                identifier,
                _workflow_model.WorkflowSpec(
                    self.sdk_workflow,
                    sub_workflows,
                ),
            )
        except _user_exceptions.FlyteEntityAlreadyExistsException:
            pass


class FlyteTask(_common_engine.BaseTaskExecutor):
    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def register(self, identifier):
        client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
        try:
            client.create_task(identifier, _task_models.TaskSpec(self.sdk_task))
        except _user_exceptions.FlyteEntityAlreadyExistsException:
            pass

    def execute(self, inputs, context=None):
        """
        Just execute the task and write the outputs to where they belong
        :param flytekit.models.literals.LiteralMap inputs:
        :param dict[Text, Text] context:
        :rtype: dict[Text, flytekit.models.common.FlyteIdlEntity]
        """
        with _common_utils.AutoDeletingTempDir("engine_dir") as temp_dir:
            with _common_utils.AutoDeletingTempDir("task_dir") as task_dir:
                with _data_proxy.LocalWorkingDirectoryContext(task_dir):
                    raw_output_data_prefix = context.get("raw_output_data_prefix", None)
                    with _data_proxy.RemoteDataContext(raw_output_data_prefix_override=raw_output_data_prefix):
                        output_file_dict = dict()

                        # This sets the logging level for user code and is the only place an sdk setting gets
                        # used at runtime.  Optionally, Propeller can set an internal config setting which
                        # takes precedence.
                        log_level = _internal_config.LOGGING_LEVEL.get() or _sdk_config.LOGGING_LEVEL.get()
                        _logging.getLogger().setLevel(log_level)

                        try:
                            output_file_dict = self.sdk_task.execute(
                                _common_engine.EngineContext(
                                    execution_id=_identifier.WorkflowExecutionIdentifier(
                                        project=_internal_config.EXECUTION_PROJECT.get(),
                                        domain=_internal_config.EXECUTION_DOMAIN.get(),
                                        name=_internal_config.EXECUTION_NAME.get(),
                                    ),
                                    execution_date=_datetime.utcnow(),
                                    stats=_get_stats(
                                        # Stats metric path will be:
                                        # registration_project.registration_domain.app.module.task_name.user_stats
                                        # and it will be tagged with execution-level values for project/domain/wf/lp
                                        "{}.{}.{}.user_stats".format(
                                            _internal_config.TASK_PROJECT.get() or _internal_config.PROJECT.get(),
                                            _internal_config.TASK_DOMAIN.get() or _internal_config.DOMAIN.get(),
                                            _internal_config.TASK_NAME.get() or _internal_config.NAME.get(),
                                        ),
                                        tags={
                                            "exec_project": _internal_config.EXECUTION_PROJECT.get(),
                                            "exec_domain": _internal_config.EXECUTION_DOMAIN.get(),
                                            "exec_workflow": _internal_config.EXECUTION_WORKFLOW.get(),
                                            "exec_launchplan": _internal_config.EXECUTION_LAUNCHPLAN.get(),
                                            "api_version": flytekit.__version__,
                                        },
                                    ),
                                    logging=_logging,
                                    tmp_dir=task_dir,
                                    raw_output_data_prefix=context["raw_output_data_prefix"]
                                    if "raw_output_data_prefix" in context
                                    else None,
                                ),
                                inputs,
                            )
                        except _exception_scopes.FlyteScopedException as e:
                            _logging.error("!!! Begin Error Captured by Flyte !!!")
                            output_file_dict[_constants.ERROR_FILE_NAME] = _error_models.ErrorDocument(
                                _error_models.ContainerError(e.error_code, e.verbose_message, e.kind)
                            )
                            _logging.error(e.verbose_message)
                            _logging.error("!!! End Error Captured by Flyte !!!")
                        except Exception:
                            _logging.error("!!! Begin Unknown System Error Captured by Flyte !!!")
                            exc_str = _traceback.format_exc()
                            output_file_dict[_constants.ERROR_FILE_NAME] = _error_models.ErrorDocument(
                                _error_models.ContainerError(
                                    "SYSTEM:Unknown",
                                    exc_str,
                                    _error_models.ContainerError.Kind.RECOVERABLE,
                                )
                            )
                            _logging.error(exc_str)
                            _logging.error("!!! End Error Captured by Flyte !!!")
                        finally:
                            for k, v in _six.iteritems(output_file_dict):
                                _common_utils.write_proto_to_file(v.to_flyte_idl(), _os.path.join(temp_dir.name, k))
                            _data_proxy.Data.put_data(
                                temp_dir.name,
                                context["output_prefix"],
                                is_multipart=True,
                            )

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def launch(
        self,
        project,
        domain,
        name=None,
        inputs=None,
        notification_overrides=None,
        label_overrides=None,
        annotation_overrides=None,
        auth_role=None,
    ):
        """
        Executes the task as a single task execution and returns the identifier.
        :param Text project:
        :param Text domain:
        :param Text name:
        :param flytekit.models.literals.LiteralMap inputs: The inputs to pass
        :param list[flytekit.models.common.Notification] notification_overrides: If specified, override the
            notifications.
        :param flytekit.models.common.Labels label_overrides:
        :param flytekit.models.common.Annotations annotation_overrides:
        :param flytekit.models.common.AuthRole auth_role:
        :rtype: flytekit.models.execution.Execution
        """
        disable_all = notification_overrides == []
        if disable_all:
            notification_overrides = None
        else:
            notification_overrides = _execution_models.NotificationList(notification_overrides or [])
            disable_all = None

        if not auth_role:
            assumable_iam_role = _auth_config.ASSUMABLE_IAM_ROLE.get()
            kubernetes_service_account = _auth_config.KUBERNETES_SERVICE_ACCOUNT.get()

            if not (assumable_iam_role or kubernetes_service_account):
                _logging.warning(
                    "Using deprecated `role` from config. "
                    "Please update your config to use `assumable_iam_role` instead"
                )
                assumable_iam_role = _sdk_config.ROLE.get()
            auth_role = _common_models.AuthRole(
                assumable_iam_role=assumable_iam_role,
                kubernetes_service_account=kubernetes_service_account,
            )

        try:
            # TODO(katrogan): Add handling to register the underlying task if it's not already.
            client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
            exec_id = client.create_execution(
                project,
                domain,
                name,
                _execution_models.ExecutionSpec(
                    self.sdk_task.id,
                    _execution_models.ExecutionMetadata(
                        _execution_models.ExecutionMetadata.ExecutionMode.MANUAL,
                        "sdk",  # TODO: get principle
                        0,  # TODO: Detect nesting
                    ),
                    notifications=notification_overrides,
                    disable_all=disable_all,
                    labels=label_overrides,
                    annotations=annotation_overrides,
                    auth_role=auth_role,
                ),
                inputs,
            )
        except _user_exceptions.FlyteEntityAlreadyExistsException:
            exec_id = _identifier.WorkflowExecutionIdentifier(project, domain, name)
        return client.get_execution(exec_id)


class FlyteWorkflowExecution(_common_engine.BaseWorkflowExecution):
    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def get_node_executions(self, filters=None):
        """
        :param list[flytekit.models.filters.Filter] filters:
        :rtype: dict[Text, flytekit.common.nodes.SdkNodeExecution]
        """
        client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
        return {
            v.id.node_id: v for v in _iterate_node_executions(client, self.sdk_workflow_execution.id, filters=filters)
        }

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def sync(self):
        """
        :rtype: None
        """
        client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
        self.sdk_workflow_execution._closure = client.get_execution(self.sdk_workflow_execution.id).closure

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def get_inputs(self):
        """
        :rtype: flytekit.models.literals.LiteralMap
        """
        client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
        execution_data = client.get_execution_data(self.sdk_workflow_execution.id)

        # Inputs are returned inline unless they are too big, in which case a url blob pointing to them is returned.
        if bool(execution_data.full_inputs.literals):
            return execution_data.full_inputs

        if execution_data.inputs.bytes > 0:
            with _common_utils.AutoDeletingTempDir() as t:
                tmp_name = _os.path.join(t.name, "inputs.pb")
                _data_proxy.Data.get_data(execution_data.inputs.url, tmp_name)
                return _literals.LiteralMap.from_flyte_idl(
                    _common_utils.load_proto_from_file(_literals_pb2.LiteralMap, tmp_name)
                )
        return _literals.LiteralMap({})

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def get_outputs(self):
        """
        :rtype: flytekit.models.literals.LiteralMap
        """
        client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
        execution_data = client.get_execution_data(self.sdk_workflow_execution.id)

        # Outputs are returned inline unless they are too big, in which case a url blob pointing to them is returned.
        if bool(execution_data.full_outputs.literals):
            return execution_data.full_outputs

        if execution_data.outputs.bytes > 0:
            with _common_utils.AutoDeletingTempDir() as t:
                tmp_name = _os.path.join(t.name, "outputs.pb")
                _data_proxy.Data.get_data(execution_data.outputs.url, tmp_name)
                return _literals.LiteralMap.from_flyte_idl(
                    _common_utils.load_proto_from_file(_literals_pb2.LiteralMap, tmp_name)
                )
        return _literals.LiteralMap({})

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def terminate(self, cause):
        """
        :param Text cause:
        """
        _FlyteClientManager(
            _platform_config.URL.get(), insecure=_platform_config.INSECURE.get()
        ).client.terminate_execution(self.sdk_workflow_execution.id, cause)


class FlyteNodeExecution(_common_engine.BaseNodeExecution):
    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def get_task_executions(self):
        """
        :rtype: list[flytekit.common.tasks.executions.SdkTaskExecution]
        """
        client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
        return list(_iterate_task_executions(client, self.sdk_node_execution.id))

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def get_subworkflow_executions(self):
        """
        :rtype: list[flytekit.common.workflow_execution.SdkWorkflowExecution]
        """
        raise NotImplementedError("Cannot retrieve sub-workflow information from a node execution yet.")

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def get_inputs(self):
        """
        :rtype: flytekit.models.literals.LiteralMap
        """
        client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
        execution_data = client.get_node_execution_data(self.sdk_node_execution.id)

        # Inputs are returned inline unless they are too big, in which case a url blob pointing to them is returned.
        if bool(execution_data.full_inputs.literals):
            return execution_data.full_inputs

        if execution_data.inputs.bytes > 0:
            with _common_utils.AutoDeletingTempDir() as t:
                tmp_name = _os.path.join(t.name, "inputs.pb")
                _data_proxy.Data.get_data(execution_data.inputs.url, tmp_name)
                return _literals.LiteralMap.from_flyte_idl(
                    _common_utils.load_proto_from_file(_literals_pb2.LiteralMap, tmp_name)
                )
        return _literals.LiteralMap({})

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def get_outputs(self):
        """
        :rtype: flytekit.models.literals.LiteralMap
        """
        client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
        execution_data = client.get_node_execution_data(self.sdk_node_execution.id)

        # Outputs are returned inline unless they are too big, in which case a url blob pointing to them is returned.
        if bool(execution_data.full_outputs.literals):
            return execution_data.full_outputs

        if execution_data.outputs.bytes > 0:
            with _common_utils.AutoDeletingTempDir() as t:
                tmp_name = _os.path.join(t.name, "outputs.pb")
                _data_proxy.Data.get_data(execution_data.outputs.url, tmp_name)
                return _literals.LiteralMap.from_flyte_idl(
                    _common_utils.load_proto_from_file(_literals_pb2.LiteralMap, tmp_name)
                )
        return _literals.LiteralMap({})

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def sync(self):
        """
        :rtype: None
        """
        client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
        self.sdk_node_execution._closure = client.get_node_execution(self.sdk_node_execution.id).closure


class FlyteTaskExecution(_common_engine.BaseTaskExecution):
    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def get_inputs(self):
        """
        :rtype: flytekit.models.literals.LiteralMap
        """
        client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
        execution_data = client.get_task_execution_data(self.sdk_task_execution.id)

        # Inputs are returned inline unless they are too big, in which case a url blob pointing to them is returned.
        if bool(execution_data.full_inputs.literals):
            return execution_data.full_inputs

        if execution_data.inputs.bytes > 0:
            with _common_utils.AutoDeletingTempDir() as t:
                tmp_name = _os.path.join(t.name, "inputs.pb")
                _data_proxy.Data.get_data(execution_data.inputs.url, tmp_name)
                return _literals.LiteralMap.from_flyte_idl(
                    _common_utils.load_proto_from_file(_literals_pb2.LiteralMap, tmp_name)
                )
        return _literals.LiteralMap({})

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def get_outputs(self):
        """
        :rtype: flytekit.models.literals.LiteralMap
        """
        client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
        execution_data = client.get_task_execution_data(self.sdk_task_execution.id)

        # Inputs are returned inline unless they are too big, in which case a url blob pointing to them is returned.
        if bool(execution_data.full_outputs.literals):
            return execution_data.full_outputs

        if execution_data.outputs.bytes > 0:
            with _common_utils.AutoDeletingTempDir() as t:
                tmp_name = _os.path.join(t.name, "outputs.pb")
                _data_proxy.Data.get_data(execution_data.outputs.url, tmp_name)
                return _literals.LiteralMap.from_flyte_idl(
                    _common_utils.load_proto_from_file(_literals_pb2.LiteralMap, tmp_name)
                )
        return _literals.LiteralMap({})

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def sync(self):
        """
        :rtype: None
        """
        client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
        self.sdk_task_execution._closure = client.get_task_execution(self.sdk_task_execution.id).closure

    @_deprecated(
        reason="Objects should access client directly, will be removed by 1.0",
        version="0.13.0",
    )
    def get_child_executions(self, filters=None):
        """
        :param list[flytekit.models.filters.Filter] filters:
        :rtype: dict[Text, flytekit.common.nodes.SdkNodeExecution]
        """
        client = _FlyteClientManager(_platform_config.URL.get(), insecure=_platform_config.INSECURE.get()).client
        return {
            v.id.node_id: v
            for v in _iterate_node_executions(
                client,
                task_execution_identifier=self.sdk_task_execution.id,
                filters=filters,
            )
        }
